:- module(_, [], [ciaobld(bundlehooks), dcg]).

:- doc(title,  "Bundle Hooks for Ciao LLVM").

:- use_module(library(bundle/bundle_flags), [get_bundle_flag/2]).
:- use_module(library(bundle/bundle_paths), [bundle_path/3]).
:- use_module(library(llists), [flatten/2]).
:- use_module(ciaobld(third_party_config), [foreign_config_atm/4]).    
:- use_module(ciaobld(messages_aux), [normal_message/2]).
:- use_module(library(process), [process_call/3]).
:- use_module(library(system), [find_executable/2]).    
:- use_module(library(file_utils), [string_to_file/2]).

% ---------------------------------------------------------------------------

:- doc(section, "Available LLVM-based tools").

% LLVM-3.9
:- bundle_flag(llvm_dir, [
    comment("Path of LLVM installation (requires 3.9)"),
    details(
      % .....................................................................
      "Specify the directory where LLVM is installed."),
    rule_default(Path, locate_llvm(Path)),
    %
    interactive
]).

:- bundle_flag(with_llvm, [
    valid_values(['yes', 'no']),
    %
    default_comment("LLVM available"),
    default_value_comment(no,
      % .....................................................................
      "LLVM has not been detected."),
    rule_default(Value, has_llvm(Value)),
    %
    interactive([advanced]),
     details(
      % .....................................................................
      "Set to \"yes\" if you wish to interface with LLVM."),
    comment("Install LLVM support")
]).

:- use_module(library(system), [file_exists/1]).

% TODO: Naive detection method, do something better
locate_llvm(Path) :-
	try_llvm_path(Path0),
	file_exists(Path0),
	!,
	Path = Path0.
locate_llvm(''). % not installed

try_llvm_path('/usr/lib/llvm-3.9'). % Ubuntu/Debian
try_llvm_path('/usr/local/Cellar/llvm/3.9.0'). % macOS (homebrew)

curr_llvm_dir(LLVMDir) :-
	LLVMDir = ~get_bundle_flag(ciao_llvm:llvm_dir),
	\+ LLVMDir = ''.

llvm_installed :- curr_llvm_dir(_LLVMDir).

has_llvm(Value) :-
	( llvm_installed -> Value = yes ; Value = no ).

with_llvm := ~get_bundle_flag(ciao_llvm:with_llvm).

% XMOS xTIMEcomposer for xC language
:- bundle_flag(with_xtime_composer, [
    valid_values(['yes', 'no']),
    %
    default_comment("xTIMEcomposer available"),
    default_value_comment(no,
      % .....................................................................
      "xTIMEcomposer has not been detected.  If you want to use the \n" ||
      "xTIMEcomposer out of the box interface it is recommended that you\n" ||
      "stop the Ciao configuration now and install xTIMEcomposer  first \n" ||
      "and/or add the directory of the xcc1llvm command line to your\n" ||
      "PATH variable."),
    rule_default(With_xTIMEcomposer, has_xtime_composer(With_xTIMEcomposer)),
    %
    interactive([advanced]),
     details(
      % .....................................................................
      "Set to \"yes\" if you wish to interface with the xTIMEcomposer.\n" ||
      "If you choose to have the xTIMEcomposer interface, you should have\n"||
      "xTIMEcomposer installed in the machine where you are compiling and\n" ||
      "using it, then and/or add to your PATH variable the directory of\n"||
      "the xcc1llvm command line."),
    comment("Install xTIMEcomposer support")
]).

m_bundle_foreign_config_tool(ciao_llvm, xtime_composer, Path) :-
	Path = ~bundle_path(ciao_llvm,'Manifest/xTIMEcomposer-config.sh').

xtime_composer_installed :-
	find_executable(~m_bundle_foreign_config_tool(ciao_llvm, xtime_composer), Exec),
	process_call(path(Exec), ['--xcc'],
	             [stdout(null), stderr(stdout), status(0)]).

has_xtime_composer(Value) :-
	( xtime_composer_installed -> Value = yes ; Value = no ).

with_xtime_composer := ~get_bundle_flag(ciao_llvm:with_xtime_composer).

% ---------------------------------------------------------------------------

:- use_module(ciaobld(builder_aux), [update_file_from_clauses/3]).

create_config_auto :-
	config_flags(Cs, []),
	File = ~bundle_path(ciao_llvm, 'src/ciao_llvm_config_auto.pl'),
	update_file_from_clauses(Cs, File, _).

config_flags -->
	[(llvm_path(_, _) :- fail)],
	% For xTIMEcomposer interface
	( { with_xtime_composer(yes) } ->
	    { foreign_config_atm(ciao_llvm, xtime_composer, 'dir', XMOSDir) },
	    { foreign_config_atm(ciao_llvm, xtime_composer, 'xpp', XPPPath) },
	    { foreign_config_atm(ciao_llvm, xtime_composer, 'xcc', XCCPath) },
	    [llvm_path(xmos, XMOSDir)],
	    [llvm_path(xpp, XPPPath)],
	    [llvm_path(xcc, XCCPath)]
	; []
	),
	% For LLVM
	( { with_llvm(yes),
	    curr_llvm_dir(LLVMDir),
	    path_concat(LLVMDir, 'bin/clang', Clang),
	    path_concat(LLVMDir, 'bin/opt', Opt)
	  } ->
	    [llvm_path(clang, Clang)],
	    [llvm_path(opt, Opt)]
	; []
	).

% ---------------------------------------------------------------------------

% Prepare source for build
% (e.g., for automatically generated code, foreign interfaces, etc.)
'$builder_hook'(prepare_build_bin) :-
	( with_llvm(yes) -> build_llvm_pass ; true ),
        create_config_auto.

:- use_module(ciaobld(cmake_aux), [b_cmake/4]).    
:- use_module(library(pathnames), [path_concat/3]).

build_llvm_pass :-
	% Build our LLVM pass using cmake.
	% We need to pass the location of LLVM at the environment.
	curr_llvm_dir(LLVMDir),
	path_concat(LLVMDir, 'lib/cmake', LLVMDirCMake),
	Env = ['LLVM_DIR'=LLVMDirCMake],
	%
	b_cmake(ciao_llvm, 'src', 'llvm2ciao', Env).

% ---------------------------------------------------------------------------
:- doc(section, "Tests and Benchmarks").

%:- use_module(ciaobld(ciaoc_aux), [runtests_dir/2]).
:- use_module(ciaobld(ciaoc_aux), [invoke_ciaosh_batch/1]).
:- use_module(library(system), [working_directory/2]).

'$builder_hook'(test) :- !,
	% runtests_dir(ciao_llvm, 'src'),
	working_directory(ThisDir, ~bundle_path(ciao_llvm, 'tests')),
	invoke_ciaosh_batch([
	  use_module(run_tests, [run/0]),
	  run
	]),
	working_directory(_, ThisDir).

