:- module(ciao_llvm, [], [assertions, regtypes, isomodes]).

:- doc(title, "Very high level interface to LLVM-based tools").
:- doc(author, "Umer Liqat").
:- doc(author, "Jose F. Morales (minor)").

:- doc(module, "This module provides a very high level interface to
   LLVM-based tools, including a compilation pass that translates LLVM
   IR into clauses.

   See code at @tt{LLVM2CIAO/} and @tt{CMakeLists.txt} for the LLVM
   pass code in C++.

   Additionally, this module also provides:
   @begin{itemize}
   @item support for other LLVM tools (XMOS xC compiler)
   @item translation from other ISA level representation (XMOS xs1 architecture)
   @end{itemize}
").

% TODO: add a model for execution

% ---------------------------------------------------------------------------

:- use_module(engine(system_info), [get_so_ext/1]).
:- use_module(library(bundle/bundle_paths), [bundle_path/3]).
:- use_module(library(process)).
:- use_module(library(pathnames), [path_concat/3]).

:- use_module(ciao_llvm(xs1_transformation)).

% ---------------------------------------------------------------------------

% (llvm_path/2 from configuration)
:- include(ciao_llvm(ciao_llvm_config_auto)).

% ---------------------------------------------------------------------------

% Get path of a LLVM-based tool (or throw exception otherwise)
get_llvm_path(Name, Path) :-
	llvm_path(Name, Path0),
	!,
	Path = Path0.
get_llvm_path(Name, _Path) :-
	throw(error(no_path_for(Name), get_llvm_path/2)).

:- use_module(ciaobld(eng_defs), [bld_cmake_path/3]).

% Path to the .so that implements the LLVM pass (note: .so in macOS and Linux)
llvm_pass_so(PassSO) :-
	bld_cmake_path(ciao_llvm, llvm2ciao, BuildDir),
	path_concat(BuildDir, 'LLVM2CIAO/libLLVM2CIAO.so', PassSO).

:- export(compile_c/2).
:- pred compile_c(+In, +Out) :: atm * atm
   # "Compile @var{In} C file into @var{Out} LLVM IR file".

% E.g., clang -S -emit-llvm -o selectionsort.ll -c benchs/selectionsort.c
compile_c(In, Out) :-
	% Get Clang command
	get_llvm_path(clang, ClangPath),
	% TODO: missing optimization level, etc.
	ClangArgs = ['-S', '-emit-llvm', '-o', Out, '-c', In],
	( process_call(ClangPath, ClangArgs, [status(0)])->
	   true
	; throw(error(failed, compile_c/2))
	).

:- export(translate_llvmIR/3).
:- pred translate_llvmIR(+LLVM, +Pl, +AssertsFile) :: atm * atm * atm
   # "Translate input @var{LLVM} LLVM-IR code into output @var{PL} clauses".

% Translate LLVM IR transformation into clauses
translate_llvmIR(LLVM, Pl, AssertsFile) :-
	get_llvm_path(opt, OPTPath),
	% Call transformation
	LLVMPassManager=OPTPath,%bundle_path(ciao_llvm, 'xcore/opt',LLVMPassManager),
	llvm_pass_so(TransformPass),
	run_llvm_pass(LLVMPassManager,TransformPass,LLVM,Pl,AssertsFile).

run_llvm_pass(LLVMPassManager,TransformPass,LLVM,Pl,AssertsFile):-
	process_call(LLVMPassManager, [
          '-analyze',
	  '-load', TransformPass,
	  '-mem2reg', '-break-crit-edges', '-loop-simplify', '-instnamer',
	  '-emit-ciao', LLVM,
	  '-of', Pl,
	  '-transform_nested_loop','yes',   % TODO: It should be a preprocess flag, this is necessary only for
	                                    % properly setting up equations in res_plai in presence of nested loops.
	  '-assrt', AssertsFile], [status(0)]),
	!.
run_llvm_pass(_LLVMPassManager,_TransformPass,_LLVM,_Pl,_AssertsFile):-
	throw(error(failed, run_llvm_pass/6)).

choose_compilation_option(o0,'-O0').
choose_compilation_option(o1,'-O1').
choose_compilation_option(o2,'-O2').

:- export(preprocess_xc/2).
% TODO: document
preprocess_xc(In, Out) :-
	get_llvm_path(xmos, XMOSPath),
	get_llvm_path(xpp, XPPPath),
	path_concat(XMOSPath,'target/include',Include),
	path_concat(XMOSPath,'target/include/gcc',IncludeGCC),
	XPPArgs = ['-version', '-x', 'xc_llvm', '-isystem', Include, '-isystem', IncludeGCC,
	           '-D__XS1B__', '-DXCC_VERSION_YEAR=12', '-DXCC_VERSION_MONTH=2',
		   '-DXCC_VERSION_MAJOR=1202', '-DXCC_VERSION_MINOR=0', '-D__XCC_HAVE_FLOAT__',
		   '-o', Out, In],
	( process_call(XPPPath, XPPArgs, [status(0)]) ->
	    true
	; throw(error(failed, preprocess_xc/2))
	).

:- export(compile_xc/4).
% Compile XC source program into some specific Layer (isa for xs1 or LLVM IR).
% Level is the optimization level (o0,o1,o2)
compile_xc(In, Out, Level, isa):- !,
	get_llvm_path(xcc, XCCPath), % Get XCC command
	choose_compilation_option(Level,CompOption),
	XCCArgs = ['-fno-subword-select', '-march=xs1b', '-quiet', '-version',
	           CompOption, '-o', Out, In],
	( process_call(XCCPath, XCCArgs, [status(0)]) ->
	    true
	; throw(error(failed, compile_xc/3))
	).
% TODO: Temporary solution, the following generates an isa compilation and not llvm.
compile_xc(In, Out, Level, llvm):- !,
	get_llvm_path(xcc, XCCPath), % Get XCC command
	choose_compilation_option(Level,CompOption),
	XCCArgs = ['-fno-subword-select', '-march=xs1b', '-quiet', '-version',
	           CompOption, '-emit-llvm', '-o', Out, In],
	( process_call(XCCPath, XCCArgs, [status(0)]) ->
	    true
	; throw(error(failed, compile_xc/3))
	),
	% Commenting out XTA debug metadata information from LLVM IR 
	% that is xcc specific and not recognized by LLVM pass manager
%	process_call(bundle_src(ciao_llvm)/xcore/'replaceXTA.sh', [Out], [status(0)]).
	process_call(path(sed), ['-ie', 's/, !xta/;, !xta/g', Out], [status(0)]),
	process_call(path(sed), ['-ie', 's/!xta/;!xta/g', Out], [status(0)]).

:- export(translate_entra_llvmIR/5).
% ISA transformation
translate_entra_llvmIR(Assembly, Pl, AssertsFile, IncludeDefaultModel, isa) :- !,
	xs1_transformation:output(Assembly, AssertsFile, IncludeDefaultModel, Pl).
%
% LLVM IR transformation (XCC) % TODO: the clang and xcc versions of translate_IR/5 to be merged.
translate_entra_llvmIR(LLVM, Pl, AssertsFile, IncludeDefaultModel, llvm) :-
	entra_llvm_path(opt, LLVMPassManager),
	entra_llvm_pass_so(TransformPass),
	run_entra_llvm_pass(LLVMPassManager,TransformPass,LLVM,Pl,AssertsFile,IncludeDefaultModel).

% TODO: ad-hoc, this should not be here
entra_llvm_path(opt, Opt) :-
	bundle_path(ciao_llvm, 'xcore/opt', Opt).

% TODO: ad-hoc, this should not be here
entra_llvm_pass_so(TransformPass) :-
	bundle_path(ciao_llvm, 'xcore/LLVMENTRA', LLVMTransformPassLib),
	get_so_ext(Ext_so),
	atom_concat(LLVMTransformPassLib, Ext_so, TransformPass).

% opt -analyze -load ../../libLLVM2CIAO.so -mem2reg -break-crit-edges -loop-simplify -instnamer -emit-ciao selectionsort.ll -of selectionsort.pl -default_model yes -pkg filename

run_entra_llvm_pass(LLVMPassManager,TransformPass,LLVM,Pl,AssertsFile,IncludeDefaultModel):-
	process_call(LLVMPassManager, [
          '-analyze',
	  '-load', TransformPass,
	  '-mem2reg', '-break-crit-edges', '-loop-simplify', '-instnamer',
	  '-emit-ciao', LLVM,
	  '-of', Pl, 
	  '-assrt', AssertsFile, 
	  '-default_model', IncludeDefaultModel], [status(0)]), % TODO: deprecate
	!.
run_entra_llvm_pass(_LLVMPassManager,_TransformPass,_LLVM,_Pl,_AssertsFile,_IncludeDefaultModel):-
	throw(error(failed, run_entra_llvm_pass/6)).

