:- module(xs1_parser,
	[extract_label/3,file_read/1,program/1],
	[assertions, nativeprops, regtypes, dcg]).

:- doc(title, "XC Assembly Parser").
:- doc(author, "Umer Liqat").

:- doc(summary, "A parser for the XS1 assembly instruction set. The
       parsed information is then used in other stages of the
       transformation from an XS1 assembly program to a Ciao
       program").

:- doc(module, "This module is a DCG-based parser of the XS1 assembly
       instruction set for XMOS devices.  It reads an assembly file
       generated for a given XC source program using the xcc
       compiler.  The module parses each line in the file and produces
       a list of instructions or procedure labels.  The resultant list
       is stored in a dynamic predicate @var{program/1}.  ").

:- use_module(library(lists)). 
:- use_module(library(dec10_io)).
:- use_module(library(write)).
:- use_module(library(strings)).
:- use_module(xs1_interpreter).

:- doc(bug, "Only a subset of instructions are added, more to be
   included in @var{inst_name} along with format in @var{inst_format}").

:- data program/1.
:- data entry_assert/2.
:- export(entry_assert/2).

%%%%%%%%%%%%%%%%%%%% Parse Assembly File %%%%%%%%%%%%%%%%%%%%%%

ignored_directives([".file", ".call", ".set", ".text", ".globl",
     ".align", ".type", ".cc_top", ".cc_bottom", ".size", ".section",
     ".typestring", ".long", ".byte", ".ascii", ".par", ".space"]).

:- data instr_address/1.
:- export(instr_address/1).
:- export(get_address/1).


%Address of instruction is automatically assigned by increment a global counter 
get_address(A):-
	instr_address(A), 
	retractall_fact(instr_address(_)),
	atom_number(A, AN),
	NewA is AN + 1, 
	atom_number(AA, NewA),
	asserta_fact(instr_address(AA)).

%:- export(file_read/1).
:- doc(file_read/1, "It reads and parse each line in the input file @var{F}. The output is stored in dynamic predicate
	@var{program/1} as a list of instructions/labels. 
   @Includedef{file_read/1}").
file_read(F):-
	retractall_fact(instr_address(_)),
	asserta_fact(instr_address('1')),
	see(F),
	(file_read_aux(R)->
	 true
	;
	    seen,fail
	),
	seen,
	data_facts:asserta_fact(program(R)),!.

%:-export(file_read_aux/1).
file_read_aux(Res):-
	get_line(X),
%	atom_codes(_XX,X),
%	(_XX='sqr_aux:'->foo;true),
	(X=end_of_file->
	 true, Res=[];
	    atom_codes(_XX,X),
	    %write(_XX),nl,
	    parseLine1(X,Result),
	    (Result='.ignore'->
	     file_read_aux(Res)
	     ;
		file_read_aux(Res1),
		append([Result],Res1,Res)
	    )
	).

process_directive(".typestring",X):-
	typestring(Symbol,Input1,Return,X,Rest),
 	Rest="",
	array_bound_args(Input1, ABA),
	append(Input1, ABA, Input),	
	(Return=[void] ->
	 Args=Input
	;
	    append(Input, [var], Args)
	),
	asserta_fact(entry_assert(Symbol,Args)),!.
process_directive(X, _R):-
	ignored_directives(L),
	member(X, L).

%num type
typeDefs(num,num,[]):-!.
%list type
typeDefs(list(A), list(T), TypeDefs):-!,
	typeDefs(A, T, TypeDefs).
%strcut type
typeDefs(X, StructName, TypeDefs):-
	functor(X, StructName, NoOfMems),
	createType(X, NoOfMems, TypeDefs).

createType(Type, Members, [TypeDef|TypeDefs]):-
	functor(Type, TypeName, Members),
	getMemberList(Type, Members, MemberNames, TypeBody, TypeDefs),
	TypeHead=..[TypeName|MemberNames],
	TypeDef=type(TypeHead, TypeBody).

getMemberList(_Type, 0, [], [],[]).
getMemberList(Type, N, [M|Ms], [B|Bs], TypeDefs):-
	arg(N, Type, (M,MType)),
	functor(MType, MTypeName, _),
	(MTypeName=list->
	 arg(1,MType, MTypeSub),
	 typeDefs(MTypeSub,_A,TypeDefs1),
	 B=..[MTypeName,M,_A]
	 ;
	    B=..[MTypeName,M],
	    typeDefs(MType,_A,TypeDefs1)
	),	    
	N1 is N-1,
	getMemberList(Type, N1, Ms, Bs, TypeDefs2),
	append(TypeDefs1, TypeDefs2, TypeDefs).

% XCC compiler adds an extra int type argument for each array argument in ISA.
array_bound_args([],[]).
array_bound_args([list(_)|Rest], [num|Args]):-
	!,
	array_bound_args(Rest, Args).
array_bound_args([_|Rest], Args):-
	array_bound_args(Rest, Args).
	
parseLine1(X,'.ignore'):-
	directive(Directive, X, Rest),
	process_directive(Directive, Rest).
 	
%emptyline
parseLine1(X,'.ignore'):- X="",!.
parseLine1(X,'.ignore'):-X=[],!.

parseLine1(X,'.ignore'):-
	ignore_line( X, _Rest),!.

parseLine1(X,Result):-
	label(R1, X, Rest),
	Rest="",
	atom_codes(Result1, R1),
	Result=..[proc_label|[Result1]],!.

parseLine1(X,Result):-
	instruction(Label,InstName1,Operands1,X,Rest),
	treat_psuedo_instr(InstName1, InstName, Operands1, Operands), 
	Rest="",
	get_address(Address),
	Operands=..[operands|Ops],
	instr_op(InstName, Ops, Format, Expanded_op, KeepExpanded),
	
	(KeepExpanded=true ->
	 Operands2=..[operands|Expanded_op],
	 Result=..[instr|[Label,Address,InstName,Format,Operands2]]
	;
	    Result=..[instr|[Label,Address,InstName,Format,Operands]]
	),!.
parseLine1(X,_):-
	atom_codes(X1, X),
	atom_concat(X1, 'not recognized', Z),
	throw(error(Z, xs1_parser:parseLine1/2)).

% Section 7.10 Assembly Programming Manual
treat_psuedo_instr(mov, add, operands(X, Y), operands(X, Y, '0')):-!.
treat_psuedo_instr(I, I, O, O):-!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- doc(instruction/3, "This DCG predicate parses a line as a string
      and extracts the elements in an instruction.  An instruction is
      composed of an optional label @var{L}, an instruction name
      @var{InstName}, and the instruction's operands @var{Operands}.
      @includedef{instruction/3}").

instruction( L, InstName, Operands)--> 
	spaces,
	labell_O(LL),{atom_codes(L,LL)},
	spaces, 
	inst_name(InstName1),{atom_codes(InstName,InstName1)},
	spaces, 
	arg_list(Operands1),{Operands=..[operands|Operands1]},!.

% begins with '.' followed by one or more letters.
%:- export(directive/3).
directive(Symbol)--> spaces, period(P), char(C), directive_aux(V), {S=[C|V], Symbol=[P|S]}.

directive_aux(Symbol)--> char(U), directive_aux(V), {Symbol=[U|V]}.
directive_aux(Symbol)--> underscore(U), directive_aux(V), {Symbol=[U|V]}.
directive_aux("")--> "".

%%%%%%%%%%%%% typestring directive %%%%%%%%%%%%%%%
%:-export(typestring/5).
%e.g.    .typestring fact, "f{si}(si,si)"
% .typestring fact, "f{si}(&(a(:si)),si)"  --> int fact(int arr[], int size)
:- export(typestring/5).
typestring(Symbol,Input,Return)-->
	spaces,
	symbol(S),
	spaces,
	comma, 
	spaces,
	typeinfo(Return, Input),
	spaces,
	{atom_codes(Symbol, S)}.
:-export(symbol/3).
symbol(Symbol)--> dollar(D), symbol_aux(V), {Symbol=[D|V]}.
symbol(Symbol)--> period(P), symbol_aux(V), {Symbol=[P|V]}.
symbol(Symbol)--> underscore(U), symbol_aux(V), {Symbol=[U|V]}.
symbol(Symbol)--> char(U), symbol_aux(V), {Symbol=[U|V]}.

symbol_aux(Symbol)--> symbol(Symbol).
symbol_aux(Symbol)--> schar(U), symbol_aux(V), {Symbol=[U|V]}.
symbol_aux(Symbol)--> cchar(U), symbol_aux(V), {Symbol=[U|V]}.
symbol_aux(Symbol)--> digit(U), symbol_aux(V), {Symbol=[U|V]}.
symbol_aux("")--> "". 

 

dollar(0'$)--> [0'$].%, {D=[0'$]}.
period(0'.)--> [0'.].
underscore(0'_)--> [0'_].

:- export(typeinfo/4).

typeinfo(Return, Input)--> "\"f{",types(Return),"}","(", types(Input),")", "\"".

%:-export(types/3).

type(T)--> "si", {T='num'}.  %signed int
type(T)--> "sl", {T='num'}.  %signed long
type(T)--> "0" , {T='void'}. %void

types(Types)--> type(T), types_aux(T_Rest),{Types=[T|T_Rest]}.
types(Types)--> "&(", derivedType(T),")", types_aux(T_Rest),{Types=[T|T_Rest]}.

types_aux(TypesList)--> ",", type(T), types_aux(TRest), {TypesList=[T|TRest]}.%append(T, TRest, TypesList)}.
types_aux(TypesList)--> ",", "&(", derivedType(T),")", types_aux(TRest), {TypesList=[T|TRest]}.%append(T, TRest, TypesList)}.
types_aux("")-->"",!.

derivedType(T) --> "a(", decimal(_D),":", type(E), ")", {T='list'(E)}.
derivedType(T) --> "a(", decimal(_D),":", derivedType(E), ")", {T='list'(E)}.

% type_transform(si, num).
% type_transform(a, list).
% type_transform(ui, num).
% type_transform(0, void).

% type(X)-->"si", {type_transform(si, X)}.
% type(X)-->"ui", {type_transform(si, X)}.
% type(X)-->"0", {type_transform(0, X)}.
% type(X)-->"&", "(",type(_Y),")",{X=_Y}.
% type(X)-->structType(X).
% type(X)-->arrayType(X).

% structType(X) --> "s","(",typename(Y),")","{",structmems(Mems),"}",!,{X=..[Y|Mems]}.
% structmems(Z) --> structmem(Y),",", structmems(X),{Z=[Y|X]}.
% structmems(X) --> structmem(Y),{X=[Y]}.
% structmem(X) --> "m","(",membername(Y),")", "{", type(_X),"}", {X=(Y,_X)}.

% arrayType(X) --> "a","(",arrsize,":",type(_Y),")",{type_transform(a, A),X=..[A,_Y]}.

% :-export(decimal/3).
% typename(X)-->symbol(XX),{atom_codes(X,XX)}.
% membername(X)-->symbol(XX),{atom_codes(X,XX)}.
% arrsize-->decimal(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

label(Label)-->symbol(Label), ":".
%:- export(ignore_line/2).
ignore_line-->
	spaces, 
	comment.
	
comment--> [0'#].

%optional label (appears as instruction label)
labell_O(L)-->label(L).
labell_O("")-->"". 

% arg_list(Args)-->
% 	spaces, 
% 	instruction_arguments_subscript_op(Args1),
% 	spaces, 
% 	comma, 
% 	spaces, 
% 	arg_list(Vs),
% 	!,
% 	{append(Args1, Vs,Args)}.

% arg_list(VV)-->
% 	instruction_arguments_subscript_op(Args1),
% 	spaces,{VV=Args1}.

arg_list(Args)-->
	spaces, 
	instruction_argument(V1),
	spaces, 
	comma, 
	spaces, 
	arg_list(Vs),
	!,
	{atom_codes(V,V1),append([V], Vs,Args)}.

arg_list(VV)-->
	instruction_argument(V1), 
	spaces,{atom_codes(V,V1),VV=[V]}.

%instruction_argument(V) --> "_", instruction_argument(D),{append("_",D,V)}.
instruction_argument(V) --> "-", instruction_argument(D),{append("-",D,V)}.
instruction_argument(V) --> digit(X), instruction_argument(D),{append([X],D,V)}.
instruction_argument(V) --> cchar(X), instruction_argument(D),{append([X],D,V)}. 
instruction_argument(V) --> schar(X), instruction_argument(D),{append([X],D,V)}.
instruction_argument(V) --> " ", instruction_argument(D),{append([32],D,V)}.
%instruction_argument(V) -->  labell(D) ,{append([32],D,V)}.
instruction_argument(V) -->  symbol(X), instruction_argument(D), {append(X,D,V)}.
instruction_argument(V) --> "[", instruction_argument(D),{append("[",D,V)}.
instruction_argument(V) --> "]", instruction_argument(D),{append("]",D,V)}.
instruction_argument("") --> "".

instruction_arguments_subscript_op(Args) -->
	instruction_argument(V1),
	"[",
	instruction_argument(V2),
	"]",
	!,
	{atom_codes(V11,V1),atom_codes(V22,V2),Args=[V11, V22]}.

extract_label(L1)-->spaces, symbol(L),{atom_codes(L1,L)}.

%:- export(hex_varr/3).

hex_varr(V) --> "-", hex_varr(D),{V=[0'-|D]}.
hex_varr(V) --> digit(X), hex_varr(D),{V=[X|D]}.
hex_varr(V) --> char(X), hex_varr(D),{V=[X|D],(X>=0'A;X>=0'a),(X=<0'F;X=<0'f)}.
%hex_varr(V) --> schar(X), hex_varr(D),{V=[X|D],X>=0'a,X=<0'f}.
hex_varr("") --> "".

%:- export(decimal/3).
decimal(N) --> digit(X), decimal(D), {N=[X|D]}.
decimal("")-->"".

spaces --> space, spaces.
spaces --> [].

comma    --> [0',].

digit(X) --> [X], {X > 47, X<58}.
schar(X) --> [X], {X>96, X<123}.
cchar(X)--> [X], {X>64,X<91}.
char(X)--> schar(X).
char(X)--> cchar(X).
space    --> [X], {X=<32}. % covers EOF, ^I, ^G, ^J, ^K, ^L, ^M, ^Z, ESC, ^_

%%%%%%%%%%%%%%%%%%%%%%% XS1 Instructions %%%%%%%%%%%%%%%%%%%%%%%

% Instruction name

% inst_name(Name)-->X, {inst_interpret(instr(_I),_A,_O,_Def,_Ref,InstName,_Type), atom_codes(InstName,X), Name=X}.
inst_name(Name)-->"entsp",{Name="entsp"}.
inst_name(Name)-->"stw",{Name="stw"}.
inst_name(Name)-->"ldw",{Name="ldw"}.
inst_name(Name)-->"ldc",{Name="ldc"}.
inst_name(Name)-->"lss",{Name="lss"}.
inst_name(Name)-->"bf",{Name="bf"}.
inst_name(Name)-->"bt",{Name="bt"}.
inst_name(Name)-->"bu",{Name="bu"}.
inst_name(Name)-->"mkmsk",{Name="mkmsk"}.
inst_name(Name)-->"retsp",{Name="retsp"}.
inst_name(Name)-->"sub",{Name="sub"}.
inst_name(Name)-->"bl",{Name="bl"}.
inst_name(Name)-->"mul",{Name="mul"}.
inst_name(Name)-->"add",{Name="add"}.
inst_name(Name)-->"eq",{Name="eq"}.
inst_name(Name)-->"zext",{Name="zext"}.
inst_name(Name)-->"ashr",{Name="ashr"}.
inst_name(Name)-->"rems",{Name="rems"}.
inst_name(Name)-->"divs",{Name="divs"}.
inst_name(Name)-->"shr",{Name="shr"}.
inst_name(Name)-->"ashr",{Name="ashr"}.
inst_name(Name)-->"ldaw",{Name="ldaw"}.
inst_name(Name)-->"maccs",{Name="maccs"}.
inst_name(Name)-->"sext",{Name="sext"}.
inst_name(Name)-->"shl",{Name="shl"}.
inst_name(Name)-->"or",{Name="or"}.
inst_name(Name)-->"ld8u",{Name="ld8u"}.

%pseudo instructions 
inst_name(Name)-->"mov",{Name="mov"}.
