:- module(xs1_transformation,
	[ 
	  transform/2,
	  output/4,
	  output_with_entry_assertion/4,
	  preds2unfold/2,
	  block_def/2,
	  get_entry_input_vars/3
	  % def/2,
	  % ref/2
	],
	[ assertions, regtypes
	]).

:- doc(title, "Graph Processing and Horn Clause transformation").
:- doc(author, "Umer Liqat").

:- doc(summary, "CFG construction at instruction and block level. The
   Graphs are then used to infer Ciao Clauses and SSA Form.").

:- doc(module, "This module construct a control flow graph from a list
   of program instructions. The control flow graph is then used to
   construct block level control flow graph. The block-cfg is
   processed further to generate horn clauses.").

:- doc(bug, "Separate Inferring block arguments, SSA Form and block
CFG construction in different modules.").

:- doc(bug, "Block names needs to match the label names for better
readability").

:- doc(bug, "Optimization required for inferring block arguments->
requires trust assertions to identify input output arguments ").

:- doc(bug, "Operand transformation of instructions manipulating
arrays produces bug in cisC_prev/4 that results in some jmp/1 literals
appearing in transformed code").

:- doc(bug, "To add module declarations").
%:- use_module(library(iso_char)).
:- use_module(xs1_parser).
:- use_module(xs1_asm_operand_util).
:- use_module(library(sets)).
:- use_module(library(sort)).
:- use_module(library(graphs/ugraphs)).
:- use_module(stack).
:- use_module(xs1_interpreter).
:- use_module(library(write)).
:- use_module(library(lists)).
:- use_module(library(aggregates)).
%:- use_module(ciaopp(preprocess_flags)).
:- use_module(library(format), [format_to_string/3]).
:- use_module(library(file_utils), [file_to_terms/2]).

:- export(l2a_map/1).
:- data l2a_map/1. %map from label to instruction address.
:- data cfg/1. % program's control flow graph
:- data call_edge/1.
:- data return_edge/1.
:- export(block/2).
:- data block/2.
:- data new_name/2.
% :- data ref/2.
% :- data def/2.


% :- export(block/2).
% :- export(prog/1).
% prog(X):-program(X), prog_aux(X).

% prog_aux([]).
% prog_aux([X|Xs]):-write(X),nl,prog_aux(Xs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% All %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- doc(transform/2, "Takes an input @var{File} and returns a ciao program representation in @var{CiaoProgram}").
transform(File, CiaoProgram):-
	getBlockForm(File,B),           %Generates a block representation in block/2

%	duplicate_shared_branch,
	
	% inst_def_ref,                    %Instruction level def/ref sets

%	getAllBlocksWithNames(B), 

	block_def_set(B),                %Infer block-level def sets. Output is in block_def/2. 

	reachableVerticesBlock,          %Fixpoint algorithm to compute all vertices to all vertices reachability

	%Infer block output arguments, uses the block defs and
	%reachablity information to infer block output vars.
	%blockoutput(b)=block_def(b) Union block_def(b') where b' is
	%reachable block from b. Output is in block_output_vars/2.
	% blockOutputVars(B),
	blockOutputVars1(B),
	updateBlockOutputVars(B),

	%It computers def_so_far for each node. This set assigns each
	%node N in the block a set of variable D where D is union of
	%all the defsets of nodes that preceds N. The result is in
	%def_so_far/2.
	computeDefUnion(B),                 

	block_ref_set(B),                %Infer block-level ref sets 

	%blockOutputVars,                 
	blockInputVars,                  %Infer block input arguments

	init_bis,                        %Initiliaze block instances set

	computeCIS,                      %Infer Current instance set for each instruction

	program_write(Program),          % Generate SSA form 

	resolve_branches(Program,Program1), % Resolve branches (e.g. branches as clauses of same predicate)
	resolve_branch_names(Program1,CiaoProgram). % Renaming of block names.

duplicate_shared_branch:-
	getAllBlocksWithNames(B),
	dsb_aux(B, Branches),
	duplicate_branches(Branches, []).

duplicate_branches([],_).

duplicate_branches([(_Name, [])|Rest],Visited):-
	duplicate_branches(Rest,Visited).

duplicate_branches([(Name, [jmp(BName)|[]])|Rest], Visited):-
	(member(BName, Visited)-> %Branch B1Name is shared previously rename and create a new block for it
	 duplicate_branch(Name, BName),
	 BName1=[];
	    BName1=[BName]
	),
	%append(B1Name1, B2Name1, BName),
	append(BName1, Visited, Visited_new),
	duplicate_branches(Rest, Visited_new).

duplicate_branches([(Name, [branch(B1Name,B2Name)|[]])|Rest], Visited):-
	(member(B1Name, Visited)-> %Branch B1Name is shared previously rename and create a new block for it
	 duplicate_branch(Name, B1Name),
	 B1Name1=[];
	    B1Name1=[B1Name]
	),
	(member(B2Name, Visited)-> %Branch B2Name is shared previously rename and create a new block for it
	 duplicate_branch(Name, B2Name),
	 B2Name1=[];
	    B2Name1=[B2Name]
	),
	append(B1Name1, B2Name1, BName),
	append(BName, Visited, Visited_new),
	duplicate_branches(Rest, Visited_new).
duplicate_branch(Name, BName):-
	block(Name, Block),
	newDuplicateBlock(BName,NewBName),
	replace_branch_in_block(Block, BName, NewBName, NewBlock),
	retractall_fact(block(Name, Block)),
	asserta_fact(block(Name, NewBlock)),
	block(BName, BNBlock),
	asserta_fact(block(NewBName,BNBlock)).
	
replace_branch_in_block(Block, BName, NewBName, NewBlock):-
	append(X,[branch(B1,B2)], Block),
	(B1 = BName->
	 NewB1 = NewBName;
	    NewB1=B1
	),
	(B2 = BName->
	 NewB2 = NewBName;
	    NewB2=B2
	),
	append(X, [branch(NewB1, NewB2)], NewBlock).

replace_branch_in_block(Block, BName, NewBName, NewBlock):-
	append(X,[jmp(B1)], Block),
	(B1 = BName->
	 NewB1 = NewBName;
	    NewB1=B1
	),
	append(X, [jmp(NewB1)], NewBlock).


dsb_aux([], []).
dsb_aux([(Name, B)|Rest], Branch):-
	dsb_aux2(B, Br),
	Block_br=(Name, Br),
	Branch=[Block_br|BrRest],
	dsb_aux(Rest, BrRest).

dsb_aux2([],[]).
dsb_aux2([branch(B1,B2)|_Bs],[branch(B1,B2)]):-!.
dsb_aux2([jmp(B1)|_Bs],[jmp(B1)]):-!.
dsb_aux2([_N|Ns], Branch):-
	dsb_aux2(Ns, Branch).
	
	
:- doc(output_with_entry_assertion/4, "Takes an assembly @var{InputFile} and writes a CiaoProgram in @var{OutputFile}, furthermore 
	entry assertion can be supplied. @var{EntryLabel} is the procedure label in the assembly and @var{Args} is the list 
        of 'int' and 'var' denoting input and output arguments respectively.").
% Example : 
% output_with_entry_assertion('./examples/fact_assembly.asm', './examples/output_fact_new.pl', 'fact', [int,var]).
output_with_entry_assertion(InputFile, OutputFile, EntryLabel, Args):-
	retractall_fact(entry_assert(_,_)),
	asserta_fact(entry_assert(EntryLabel,Args)),
	output(InputFile,nil,yes,OutputFile).

:- doc(output/3, "Takes an assembly @var{InputFile} and writes a CiaoProgram in @var{OutputFile}. The @var(AssertsFile) is an optional file (created automatically by the frontend transformation of xc assertions) path to check assertions that are inserted into the output. This ").
output(InputFile, AssertsFile, _IncludeDefaultModel, OutputFile):-
	retractall_fact(entry_assert(_,_)),
	transform(InputFile,CiaoProgram),
	open(OutputFile,write,Stream),
	write_file_header(_IncludeDefaultModel, Stream),
	write_frontend_assrt(Stream, AssertsFile),
	push_warning_flags(Stream),
	write2file(CiaoProgram, Stream),
	pop_warning_flags(Stream),
	close(Stream).
write_file_header(IncludeDefaultModel, Stream):-
	%atom_concat(':-module(_,[',ExportPredicate,Module1),
	
	% atom_concat(':-module(_,[','',Module1),
	% atom_concat(Module1, '],[ciaopp_xcore(model/instructions)', Module2),
	% ( current_pp_flag(xc_energy_cost, yes)
	% -> atom_concat(Module2, ',ciaopp_xcore(model/energy)',Module3)
	%  ; Module3 = Module2
	% ),
	% atom_concat(Module3, ']).',Module),

	%Module = ':- module(_1,[],[assertions,regtypes,nativeprops]).',
	write(Stream,':- module(_1,[],[assertions,regtypes,nativeprops]).'),
	nl(Stream),
	(IncludeDefaultModel=yes->
	 write(Stream,':-use_package(ciaopp_xcore(model/instructions)).'),
	 nl(Stream),
	 write(Stream,':-use_package(ciaopp_xcore(model/energy)).'),
	 nl(Stream)
	;
	    true
	),	
	write_entry_assrts(Stream).

write_frontend_assrt(Stream, AssertsFile):-
	file_to_terms(AssertsFile, FrontEndAsserts),
	write_frontend_assrt_aux(Stream, FrontEndAsserts).
write_frontend_assrt(_,_).

write_frontend_assrt_aux(_Stream, []).
write_frontend_assrt_aux(Stream, [FEA|FEAs]):-
	write(Stream, FEA),
	write(Stream, '.'),
	nl(Stream),
	write_frontend_assrt_aux(Stream, FEAs).
	
write_entry_assrts(Stream):-
	findall((Label,X), entry_assert(Label, X), Labels),
	write_entry_assrts_aux(Labels, Stream).

write_entry_assrts_aux([], _Stream).
write_entry_assrts_aux([(EntryLabel, Args)|Labels], Stream):-
	length(Args, L1),
	atom_number(L, L1),
	atom_concat(EntryLabel,'/',EntryLabel2),
	atom_concat(EntryLabel2,L,ExportPredicate),
	atom_concat(':-entry ', ExportPredicate, Entry1),
	atom_concat(Entry1, ' : ', Entry2),
	get_entry_args(Args, EntryArgs),
	atom_concat(Entry2, EntryArgs, Entry3),
	atom_concat(Entry3, '.', EntryAssertion),
	write(Stream, EntryAssertion),
	nl(Stream),
	write_entry_assrts_aux(Labels, Stream).



get_entry_args([], '').
get_entry_args([Arg], Arg).
get_entry_args([A,B|Rest], EntryArgs) :-
	get_entry_args([B|Rest], E),
	
	format_to_string("~q", A, S),
	atom_codes(A1, S),

	atom_concat(A1, ' * ', AStar),
	atom_concat(AStar, E, EntryArgs).

push_warning_flags(Stream):-
	nl(Stream),
	write(Stream, ':- push_prolog_flag(discontiguous_warnings, off).'),
	nl(Stream),
	write(Stream, ':- push_prolog_flag(single_var_warnings, off).'),
	nl(Stream),
	write(Stream, ':- push_prolog_flag(unused_pred_warnings, no).'),
	nl(Stream).

pop_warning_flags(Stream):-
	nl(Stream),
	write(Stream, ':- pop_prolog_flag(discontiguous_warnings).'),
	nl(Stream),
	write(Stream, ':- pop_prolog_flag(single_var_warnings).'),
	nl(Stream),
	write(Stream, ':- pop_prolog_flag(unused_pred_warnings).'),
	nl(Stream).

	
write2file([],_).
write2file([Clause|Rest],Stream):-
	Clause=[Head|Body],
	nl(Stream), 
	write(Stream,Head),
	write(Stream,':-'),
	nl(Stream),
	write2file_aux(Body,Stream),
	write2file(Rest,Stream).
write2file_aux([],_).
write2file_aux([Literal|Rest],Stream):-
	write(Stream, Literal),
	(Rest=[]->
	 write(Stream,'.');
	    write(Stream,',')
	),
	nl(Stream),
	write2file_aux(Rest,Stream).
	
	
	
:- data branch_table/2.
%:- export(branch_table/2).
%:- export(resolve_branches/2).
%:- export(resolve_branch_names/2).
%resolve_branch


% resolve_branch_names([ (BlockName, Input, Output) ],[ ]).
resolve_branch_names([],[]).
resolve_branch_names([[(BlockName,_I,_O)|R]|Rest],[[NewName|R]|NRest]):-
	branch_new_name(BlockName,NewName),
	resolve_branch_names(Rest,NRest).
%:- data bb_postprocessing/2.
%branch_block_postprocessing():-

branch_new_name(BName,NewName):-
	block_input_vars(BName,Input1),
	block_output_vars(BName,Output1),
	block(BName,Body),
	append(_X,[Last|[]],Body),
	( (branch_table(branch(BName,BName2),NewName1);branch_table(branch(BName2,BName),NewName1))
	-> 
	block_input_vars(BName2,Input2),
	block_output_vars(BName2,Output2),
	% (AUB)-(B-A)    
	ord_union(Input1,Input2,Input_U),
	ord_subtract(Input2,Input1, Input_S),
	ord_subtract(Input_U, Input_S, Input_US),
	compile_input_args(Input_U, Input_S, Input_US, Input),
	ord_union(Output1,Output2,Output),
	NewName2=NewName1;
%	getBranchingVar_value(InputX, BName, Input, _Bvar);          % branching variable set to 0/1

	    NewName2=BName,
	    Output=Output1,
	    toVars(Input1,Input)
	),
	currInst_vars_b(BName, Last, Output, Output3),
	append(Input,Output3,Args),
	NewName=..[NewName2|Args].


compile_input_args([], _Input_S, _Input_US, []).
compile_input_args([V|Vs], Input_S, Input_US, [U|Input]):-
	member(V, Input_S),
	toVars([V],U_list),
	U_list=[U1],
	atom_concat('', U1, U),
	compile_input_args(Vs, Input_S, Input_US, Input).
compile_input_args([V|Vs], Input_S, Input_US, [U|Input]):-
	member(V, Input_US),
	toVars([V],U_list),
	U_list=[U],
	compile_input_args(Vs, Input_S, Input_US, Input).
	
	
:- data branch_literal/3.

get_branch_var(Name, B, [Bvar]):-
	B=[X|_R],
	branch_info_table(_BlockName, _A,BranchType,Bvar,T1,T2), %T1 -> if jump successful, T2-> if not successful
	(T1=X->  %BName reached with successful branch jump
	 getBranch_var_val(BranchType,I,true),
	 asserta_fact(branch_literal(Name, Bvar, 1));
	    T2=X,  %BName is reached with non-successful branch jump
	    getBranch_var_val(BranchType,I,false),
	    asserta_fact(branch_literal(Name, Bvar, 0))
	), !.
get_branch_var(_Name, _B, []).

% getBranchingVar_value(InputU, BName, Input_val, Bvar):-
% 	block(BName, Body),
% 	Body=[X|_R], 
% 	branch_info_table(_A,BranchType,Bvar,T1,T2), %T1 -> if jump successful, T2-> if not successful
% 	(T1=X->  %BName reached with successful branch jump
% 	 getBranch_var_val(BranchType,I,true);
% 	    T2=X,  %BName is reached with non-successful branch jump
% 	    getBranch_var_val(BranchType,I,false)
% 	),
% 	InputU=[_I|InputU_aux],
% 	Input_val=[I|InputU_aux].

getBranch_var_val(bf,'0',true).
getBranch_var_val(bf,'1',false).
getBranch_var_val(bt,'1',true).
getBranch_var_val(bt,'0',false).
	
currInst_vars_b(_,_,[],[]).
currInst_vars_b(BName,Node,[O|Os],[NO|NOs]):-
	cisC_prev(BName,Node,O,NO),!,
	currInst_vars_b(BName,Node,Os,NOs).
currInst_vars_b(BName,Node,[O|Os],[NO|NOs]):-    %when combining the arguments of the clause heads for the same predicate(branch)
	xs1_op2ciao_op(O,NO),
	currInst_vars_b(BName,Node,Os,NOs).

toVars([],[]).
toVars([O|Os],[NO|NOs]):-
	xs1_op2ciao_op(O,NO),
	toVars(Os,NOs).
toVars(['0'|Os],['0'|NOs]):-
	toVars(Os,NOs).
toVars(['1'|Os],['1'|NOs]):-
	toVars(Os,NOs).
	
%Pre-req program_write(Program)	
resolve_branches([],[]).
resolve_branches([Clause|Rest],[NewClause|NewRest]):-
	Clause=[(BlockName,_,_)|_R],
	resolve_branches_aux(BlockName,Clause,NewClause),
	resolve_branches(Rest,NewRest).

resolve_branches_aux(_,[],[]).
%resolve_branches_aux(BName, [branch(B1,B2)|Rest], NewRest):- 
%	branch_table(branch(B1,B2),_),                 %Already processed
%	resolve_branches_aux(BName, Rest, NewRest).
	 
resolve_branches_aux(BName, [branch(B1,B2)|Rest], [NewClause|NewRest]):- !,
	atom_concat(B1,B2,BNew),
	block_input_vars(B1,Input1),
	block_input_vars(B2,Input2),
	ord_union(Input1,Input2,Input),
	block_output_vars(B1,Output1),
	block_output_vars(B2,Output2),
	ord_union(Output1,Output2,Output),
	currInst_vars(BName, branch(B1,B2), Output, NewOutput),
	currInst_vars2(BName, branch(B1,B2), Input, NewInput),       %For input arguments, latest instances of last node are required.
	append(NewInput,NewOutput,Args),
	NewClause=..[BNew|Args],
	retractall_fact(branch_table(branch(B1,B2),_)),
	asserta_fact(branch_table(branch(B1,B2),BNew)),
	resolve_branches_aux(BName, Rest, NewRest).

% resolve_branches_aux(BName, [ calll(B1, _CallCounter)|Rest], [NewClause|NewRest]):- !,
% 	block_input_vars(B1,Input),
% 	block_output_vars(B1,Output),
% 	currInst_vars(BName, calll(B1, _CallCounter), Output,NewOutput),
% 	currInst_vars2(BName, calll(B1, _CallCounter), Input,NewInput),
% 	append(NewInput,NewOutput,Args),
% 	NewClause=..[B1|Args],
% 	resolve_branches_aux(BName, Rest, NewRest).


	
resolve_branches_aux(N,[L|Rest],[L|NewRest]):-
	resolve_branches_aux(N,Rest,NewRest).

currInst_vars(_,_,[],[]).
currInst_vars(BName,Node,[O|Os],[NO|NOs]):-
	cisC_prev(BName,Node,O,NO),
	currInst_vars(BName,Node,Os,NOs).

currInst_vars2(_,_,[],[]).
currInst_vars2(BName,Node,[O|Os],[NO|NOs]):-
	block(BName,Body),
	getPrevNode(Node,PrevNode,Body),
 	( cisC_prev(BName,PrevNode,O,NO)->
	  true;
	    (entry_assert(BName, _)->
	     xs1_op2ciao_op(O,NO);
	     fail
	    )
	),
 	currInst_vars2(BName,Node,Os,NOs).

getPrevNode(Node,PrevNode,[PrevNode,Node|_Ns]):-!.
getPrevNode(Node,non,[Node|_Ns]):-!.
getPrevNode(Node,PrevNode,[_N|Ns]):-
	getPrevNode(Node,PrevNode,Ns).
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% para SSA %%%%%%%%%%%%%%%%%%%%%%%%%
%Block Instance Store(BIS)
%:- export(bis/3).
:- data bis/3.
%:-export(init_bis/0).
%getBlockForm('~/transform/fact_assembly.txt',Blocks), inst_def_ref, getAllBlocksWithNames(B),block_def_set(B),reachableVerticesBlock,blockOutputVars,computeDefUnion,block_ref_set(B),blockOutputVars, blockInputVars.
%initializes bis for each block with the input variables to the block.
init_bis:-
	retractall_fact(bis(_,_,_)),
	getAllBlocksWithNames(B),
	init_bis_aux(B).
init_bis_aux([]).
init_bis_aux([(Name,_B)|Bs]):-
	block_input_vars(Name,Vars),
	init_bis_aux1(Name, Vars),
	init_bis_aux(Bs).
init_bis_aux1(_N,[]).
% init_bis_aux1(Name, [V|Vs]):-
% 	asserta_fact(bis(Name,V,V)),
% 	init_bis_aux1(Name,Vs).
init_bis_aux1(Name, [V|Vs]):-
	asserta_fact(bis(Name,V,0)),
	init_bis_aux1(Name,Vs).
%CurrInst 
%:-export(currInst/3).
% currInst(BlockName,Var,CurrVarr):-
% 	bis(BlockName,Var,CurrVarr).
currInst(BlockName,Var,CurrVarr):-
	bis(BlockName,Var,Counter),
	(Counter=0->
	 CurrVarr=Var;
	    atom_number(NCounter, Counter),
	    atom_concat(Var, '-',NVar),
	    atom_concat(NVar, NCounter, CurrVarr)
	).
%NewInst
%:- export(newinst/3).

% newinst(BlockName,Var,NewVarInstance):-
% 	(bis(BlockName,Var,VarInst)-> %if
% 	 retract_fact(bis(BlockName,Var,VarInst)),
% 	 atom_concat(VarInst,'-1',NewVarInstance),
% 	 asserta_fact(bis(BlockName,Var,NewVarInstance))
% 	;  %else
% 	    asserta_fact(bis(BlockName,Var,Var)),
% 	    NewVarInstance=Var,
% 	    block(BlockName,_)
% 	).

newinst(BlockName,Var,NewVarInstance):-
	(bis(BlockName,Var,VarInst)-> %if
	 retract_fact(bis(BlockName,Var,VarInst)),
	 VarInst2 is VarInst + 1,
	 atom_number(VarInst1,VarInst2),
	 atom_concat(Var,'-',NewVarInstance1),
	 atom_concat(NewVarInstance1,VarInst1,NewVarInstance),
	 asserta_fact(bis(BlockName,Var,VarInst2))
	;  %else
	    asserta_fact(bis(BlockName,Var,0)),
	    NewVarInstance=Var,
	    block(BlockName,_)
	).
:- data duplicate_block/2.
newDuplicateBlock(BlockName,NewBlockName):-
	atom_concat(BlockName, '_', BlockName1),
	(duplicate_block(BlockName, Counter)-> %if
	 retract_fact(duplicate_block(BlockName, Counter)),
	 Counter1 is Counter + 1,
	 atom_number(VarInst2,Counter1),
	 atom_concat(BlockName1, VarInst2, NewBlockName),
	 asserta_fact(duplicate_block(BlockName,Counter1))
	;  %else
	    asserta_fact(duplicate_block(BlockName,0)),
	    atom_concat(BlockName1, '0', NewBlockName)
	).

%baseInstance
%:-export(baseInstance/2).
% baseInstance(Var, BaseInstance):-
% 	atom_concat(NewVar,'-1',Var),
% 	baseInstance(NewVar, BaseInstance).
% baseInstance(Var, BaseInstance):-
% 	\+atom_concat(_BaseInstance1,'-1',Var),
% 	BaseInstance=Var.

%Current Instance Set(CIS)
% same as bis but it stores current instances at node level unlike bis which keeps track of current instances at block level
:- export(cis/4).
:- export(cisC/4).
:- export(cisC_prev/4).
:- data cis_prev/4.

cisC(BlockName,Node,Var,NewVarInst):-
	cis(BlockName,Node,Var,NewVarInst1),
	xs1_op2ciao_op(NewVarInst1,NewVarInst).


% previous instances of first node in block BlockName
cisC_prev(BlockName,non,Var,NewVarInst):-
  	cis(BlockName,BlockName,Var,NewVarInst1),!,
	xs1_op2ciao_op(NewVarInst1,NewVarInst).

% instances of PrevNode in the block BlockName
cisC_prev(BlockName,PrevNode,Var,NewVarInst):- 
  	cis(BlockName,PrevNode,Var,NewVarInst1),!,
	xs1_op2ciao_op(NewVarInst1,NewVarInst).


:- data cis/4.
%:-export(computeCIS/0).
%getBlockForm('~/transform/fact_assembly.txt',Blocks), inst_def_ref, getAllBlocksWithNames(B),block_def_set(B),reachableVerticesBlock,blockOutputVars,computeDefUnion,block_ref_set(B),blockOutputVars, blockInputVars,init_bis.
computeCIS:-
	retractall_fact(cis(_,_,_,_)),
	getAllBlocksWithNames(B),
	computeCIS_aux(B),!.

computeCIS_aux([]).
computeCIS_aux([(BlockName,B)|Bs]):-
	block_input_vars(BlockName, Vars),
	currInstSet(BlockName, BlockName, Vars), %cis_at_block
%	cis_till_node(BlockName, B), % the empty list will be set of input variables in case we know in advance
	cis_till_node1(BlockName, B, Vars),
	computeCIS_aux(Bs).



% Def set at node n is the union of all the def sets of previous nodes.

cis_till_node1(_BlockName,[],_DefSoFar).

% cis_till_node1(BlockName,[calll(X, _CallCounter)|Ns],DefSoFar1):-
% 	block_output_vars(X,DefSet),
% 	def_so_far(BlockName-calll(X, _CallCounter),DefSoFar2),
% 	ord_union(DefSoFar1,DefSoFar2,DefSoFar),
% 	ord_intersection_diff(DefSoFar,DefSet,Intersect,NotIntersect),
% 	currInstSet(BlockName,calll(X, _CallCounter),NotIntersect),                       %save currinst set
% 	newInstSet(BlockName,calll(X, _CallCounter),Intersect),
% 	ord_union(Intersect,NotIntersect,DefSoFarNew),
% 	cis_till_node1(BlockName,Ns,DefSoFarNew).

cis_till_node1(BlockName,[jmp(X)|Ns],DefSoFar1):-
	block_output_vars(X,DefSet),
	def_so_far(BlockName-jmp(X),DefSoFar2),
	ord_union(DefSoFar1,DefSoFar2,DefSoFar),
	ord_intersection_diff(DefSoFar,DefSet,Intersect,NotIntersect),
	currInstSet(BlockName,jmp(X),NotIntersect),                       %save currinst set
	newInstSet(BlockName,jmp(X),Intersect),
	ord_union(Intersect,NotIntersect,DefSoFarNew),
	cis_till_node1(BlockName,Ns,DefSoFarNew).

cis_till_node1(BlockName,[branch(X,Y)|Ns],DefSoFar1):-
	block_output_vars(X,OutputVarX),
	block_output_vars(Y,OutputVarY),
	ord_union(OutputVarX,OutputVarY,DefSet),
	def_so_far(BlockName-branch(X,Y),DefSoFar2),
	ord_union(DefSoFar1,DefSoFar2,DefSoFar),
	ord_intersection_diff(DefSoFar,DefSet,Intersect,NotIntersect),
	currInstSet(BlockName,branch(X,Y),NotIntersect),                   %save currinst set
	newInstSet(BlockName,branch(X,Y),Intersect),
	ord_union(Intersect,NotIntersect,DefSoFarNew),
	cis_till_node1(BlockName,Ns,DefSoFarNew).

cis_till_node1(BlockName,[N|Ns],DefSoFar1):-
	def(% BlockName-
	   N,DefSet1),
	sort(DefSet1, DefSet),
	def_so_far(BlockName-N,DefSoFar2),
	ord_union(DefSoFar1,DefSoFar2,DefSoFar),
	ord_intersection_diff(DefSoFar,DefSet,Intersect,NotIntersect),
	currInstSet(BlockName,N,NotIntersect),                             %save currinst set
	newInstSet(BlockName,N,Intersect),                                 %save new instances of already defined variables
	ord_union(Intersect,NotIntersect,DefSoFarNew),
	cis_till_node1(BlockName,Ns,DefSoFarNew).








currInstSet(_Name, _Node, []).
currInstSet(Name, Node, [V|Vs]):-
	currInst(Name,V,CurrVar),
	retractall_fact(cis(Name,Node,V,CurrVar)),
	asserta_fact(cis(Name,Node,V,CurrVar)),
	currInstSet(Name,Node,Vs).
newInstSet(_Name, _Node, []).
newInstSet(Name, Node, [V|Vs]):-
	newinst(Name,V,CurrVar),
	retractall_fact(cis(Name,Node,V, CurrVar)),
	asserta_fact(cis(Name,Node,V,CurrVar)),
	newInstSet(Name,Node,Vs). 
% newinstance(Block,Var,NewInstance):-
% 	bis(Block,BIS),
	



% registers([r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11]).
% %:-export(register2var/2).
% register2var(X,Z):-
% 	registers(R),
% 	member(X,R),
% 	Z=X.
% %:-export(sp/2).
% sp(X,Z):-
% 	X=sp(Y),
% 	num(Y),
% 	Y>=0,
% 	Z=sp-Y.
% %:-export(dp/1).
% dp(X):-
% 	X=dp(Y),
% 	num(Y),
% 	Y>=0.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%:- data reachable_set/2.
%get_reachable_set(Start,
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% new_name(bu,bu).
% new_name(bf,bf).
% new_name(bt,bt).
% new_name(bl,bl).

% makeVariable(r0,'R0').
% makeVariable(_R,_R).

:- export(cleanup/0).
cleanup:-
	data_facts:retractall_fact(entry_assert(_, _)),
	data_facts:retractall_fact(branch_literal(_, _, _)),
	data_facts:retractall_fact(branch_table(_,_)),
	data_facts:retractall_fact(branch_info_table(_,_,_,_, _, _)),
	data_facts:retractall_fact(bis(_,_,_)),
	data_facts:retractall_fact(cis(_,_,_,_)),
	% data_facts:retractall_fact(def(_,_)),
	% data_facts:retractall_fact(ref(_,_)),
	data_facts:retractall_fact(block_def(_,_)),
	data_facts:retractall_fact(block_ref(_,_)),
	data_facts:retractall_fact(program(_)),
	data_facts:retractall_fact(l2a_map(_)),
	data_facts:retractall_fact(cfg(_)),
	data_facts:retractall_fact(call_edge(_)),
	data_facts:retractall_fact(return_edge(_)),
	data_facts:retractall_fact(block(_,_2)),
	data_facts:retractall_fact(blockName(_)),
	data_facts:asserta_fact(blockName(0)),
	data_facts:retractall_fact(duplicate_block(_,_)),
	clearstack.

%dynamic f/1.

% registers(Registers):- 
% 	Registers=[r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11].
% isRegister(R):-
% 	registers(Regs),
% 	member(R,Regs).

% Call from node V to W, and returns to X
:- export(callEdge/3).
callEdge(V,W,X):-
	data_facts:current_fact(call_edge(V-W)),
	data_facts:current_fact(return_edge(V-X)).

%:- export(removeCallEdge/3).
% removeCallEdge(V,W,X):-
% 	data_facts:retract_fact(call_edge(V-W)),
% 	data_facts:retract_fact(return_edge(V-X)).

%:- export(outEdges/3).
outEdges(V, Graph,  OutEdgeTargets):-
	neighbors(V,Graph,OutEdgeTargets).

%:- export(inEdges/3).
inEdges(V, Graph,  InEdgeSources):-
	point_to(V,Graph,InEdgeSources).

%:- export(getAllBlocks/1).
% getAllBlocksNames(Names):-
% 	findall(Name,block(Name,_Block),Names).
%:- export(getAllBlocksNames/1).
%:- export(getAllBlocks/1).
%getAllBlocks(Blocks):-
%	findall(Block,Name^block(Name,Block),Blocks).	
:- export(getAllBlocksWithNames/1).
getAllBlocksWithNames(Blocks):-
%	findall(N,block(N,_Block),Ns),
%	findall(B,block(_Name,B),Bs),
	findall((Name,Block),block(Name,Block),Blocks).



%%%%%%%%%%%%%%%%%%%% Output %%%%%%%%%%%%%%%%%%%%
% :- export(file_write/1).
% :- doc(file_write/1, "Takes an ouput @var{File} and writes the block representation in it").
%file_write('~/transform/fact_assemblyOutput.pl').
%getBlockForm('~/transform/fact_assembly.txt',Blocks), inst_def_ref, getAllBlocksWithNames(B),block_def_set(B),reachableVerticesBlock,blockOutputVars,computeDefUnion,block_ref_set(B),blockOutputVars, blockInputVars,init_bis,computeCIS.


%:-export(program_write/1).
%:- doc(program_write/1, "Retrun horn clause predicates of the current file and writes the block representation in it").
program_write(Program):-
         findall((Name,Block),block(Name,Block),Blocks),
	 program_write_aux1(Blocks,Program).

program_write_aux1([],[]).
program_write_aux1([(Name,Block)|Blocks],[Program6|Program]):-
	block_input_vars(Name,InputVars),
	block_output_vars(Name,OutputVars),
	Program5=(Name,InputVars,OutputVars),
	program_write_block(Name, Block,BlockProgram1,non,BranchLit),
	append(BranchLit, BlockProgram1, BlockProgram),
	%BlockProgram=[BranchLit|BlockProgram1],
	Program6=[Program5|BlockProgram],
	program_write_aux1(Blocks,Program).

program_write_block(_Name, [],[],_PrevNode,[]):-!.
program_write_block(_Name, [],[],_PrevNode,_Lit).
program_write_block(Name, [Node|Rest],Program,PrevNode, Lit):-
	getInstrFromProgram(Node,Instr),
	interpret(Name, Node, Instr,Literal,PrevNode),
	branch_lit(Name, Node, Lit),
	program_write_block(Name, Rest,Program1,Node, Lit),
	append(Literal, Program1, Program),!.

branch_lit(_BlockName, Node, [Lit]):-
	branch_info_table(_, _V1, Branch, Arg, Node, _V2),
	xs1_op2ciao_op(Arg, Arg1),
	(Branch='bt'->
	 atom_concat(Arg1,'\\=0',Lit)
	; %Branch=bf since only two kinds of branches bf or bt
	    atom_concat(Arg1, '=0', Lit)
	),!.

branch_lit(_BlockName, Node, [Lit]):-
	branch_info_table(_, _V1, Branch, Arg, _A, Node),
	xs1_op2ciao_op(Arg, Arg1),
	(Branch='bt'->
	 atom_concat(Arg1,'=0',Lit)
	; %Branch=bf since only two kinds of branches bf or bt
	    atom_concat(Arg1, '\\=0', Lit)
	),!.

branch_lit(_BlockName, _Node, _Lit).



% writeArgs_p([],' ').
% writeArgs_p([V|Vs], Vars):-
% 	atom2var(V,V1),
% 	writeArgs_p(Vs,Vars2),
% 	(Vs=[]->
% 	 atom_concat(V1,Vars2,Vars);
% 	    atom_concat(V1,',',Vars1),
% 	    atom_concat(Vars1,Vars2,Vars)
% 	).
	
%:- export(getInstrFromProgram/2).
% getInstrFromProgram(Node,instr(Label,Node,Instr,Format,Operands)):-
% 	program(P),
% 	member(instr(Label,Node,Instr,Format,Operands),P).
% getInstrFromProgram(call(X),call(X)).

%%%%%%%%%%%%%%%%%%%%% Block Form %%%%%%%%%%%%%%%$

:- export(getBlockForm/2).
:- doc(getBlockForm/2, "Takes an assembly @var{File} as an input and generates @var{Blocks} representation.").
% getBlockForm(Blocks,'~/transform/fact_assembly.txt').
getBlockForm(File, Blocks1):-
	cleanup,
	file_read(File),!,
	generate_graph(G,_F), 
	blockGraph(G,_Blocks),     % List of blocks
%	addCallNode,                % Add call nodes for all call sites in all blocks in database block/2
%	mergeBlocks(G),             % Merge the return block with the call block. It happens when the branch is bl
%	renameBlocks(G),            % Rename two blocks originating from the same branch.
	addJmpBranchNodes(G),            % Add branch/jmp nodes at the end of each block if the last instr is a jump.
	getAllBlocksWithNames(Blocks1).

%%%%%%%%%%%%%%%%  Block-def  %%%%%%%%%%%%%%%%%%%
:- data loc_counter/1.
loc_counter(0).
:- export(get_call_counter/1).
get_call_counter(X):-
	loc_counter(N),
	X is N + 1,
	retractall_fact(loc_counter(_)),
	asserta_fact(loc_counter(X)).
	

%%%%%%%%%%%%%%%%%%%% Output %%%%%%%%%%%%%%%%%%%%
% %:- export(inst_def_ref/0).
% % :- doc(file_write/1, "Takes an ouput @var{File} and writes the block representation in it").
% %file_write('~/transform/fact_assemblyOutput.pl').
% inst_def_ref:-
% %	 open(File,write,Stream), 
%          findall((Name,Block),block(Name,Block),Blocks),
% 	 inst_def_ref_aux1(Blocks).

% inst_def_ref_aux1([]).
% inst_def_ref_aux1([(Name,Block)|Blocks]):-
% 	process_block(Name, Block),
% 	inst_def_ref_aux1(Blocks).

% process_block(_N, []).
% process_block(Name, [Node|Rest]):-
% 	getInstrFromProgram(Node,Instr),
% 	write(Node),nl,
% 	inst_interpret(Instr,A,_Operands, Def, Ref, _Inst,_Type),
% 	(A=nil->
% 	 true;
% 	    data_facts:asserta_fact(def(% Name-
% 				       A,Def)),
% 	    data_facts:asserta_fact(ref(% Name-
% 				       A,Ref))
% 	),
% 	process_block(Name, Rest).
% % process_block(Name, [Node|Rest]):-
% % 	getInstrFromProgram(Node,Instr),
% % 	inst_interpret(Name,_,Instr,_Literal,_,_,_),         %Also creates def/ref sets as dynamic facts
% % 	process_block(Name, Rest).
:-export(getInstrFromProgram/2).
def(Node, DefSet):-
	getInstrFromProgram(Node,Instr),
	inst_interpret(Instr,Node,_Operands, DefSet, _RefSet, _Inst,_Type).

ref(Node, RefSet):-
	getInstrFromProgram(Node,Instr),
	inst_interpret(Instr,_A,_Operands, _DefSet, RefSet, _Inst,_Type).
	
%:- export(getInstrFromProgram/2).
getInstrFromProgram(Node,instr(Label,Node,Instr,Format,Operands)):-
	program(P),
	member(instr(Label,Node,Instr,Format,Operands),P),!.
%getInstrFromProgram(calll(_X, _CallCounter),calll(_X, _CallCounter)).
getInstrFromProgram(jmp(_X),jmp(_X)).
getInstrFromProgram(branch(_X,_Y),branch(_X,_Y)).
%:- export(block_def/2).
%:- export(block_ref/2).
:- data block_def/2.
:- data block_ref/2.
%:- export(block_def_set/1).
%:- data entry_assert/2.
%entry_assert('fact',[int,var]).


get_entry_input_vars(Label,In, Out):-
	entry_assert(Label,Args),
	geiv_aux(Args,In, Out,0,0).

geiv_aux([],[],[],_,_).
geiv_aux([num|As],[I|Is],Out,N, M):-
	atom_number(N2, N),
	atom_concat('r',N2,I),
	N1 is N + 1,
	geiv_aux(As,Is,Out,N1, M).

geiv_aux([list(_)|As],[I|Is],Out,N, M):-
	atom_number(N2, N),
	atom_concat('r',N2,I),
	N1 is N + 1,
	geiv_aux(As,Is,Out,N1, M).
geiv_aux([var|As],In,[O|Os],N, M):-
	atom_number(M2, M),
	atom_concat('r',M2,O),
	M1 is M + 1,
	geiv_aux(As,In,Os,N,M1).

block_def_set([]).
block_def_set([(Name,Block)|Blocks]):- %function  block
	entry_assert(Name,_),
	getUnionOfDefs_From_Block(Name, Block,[],DefSetUnion1),
	filter_args(0, Name, DefSetUnion1, DefSetUnion),
	(get_entry_input_vars(Name, _In, Out)->   %Temporary fix won't work for multiple entry points.
	 filter_args_aux2(DefSetUnion, Out, DefSetUnion2),!
	;
	    DefSetUnion2=[],!
	),
	asserta_fact(block_def(Name,DefSetUnion2)),
	block_def_set(Blocks).

block_def_set([(Name,Block)|Blocks]):- %intermediate block
	\+ entry_assert(Name,_),!,
	getUnionOfDefs_From_Block(Name, Block,[],DefSetUnion),
	% filter_args(0, Name, DefSetUnion1, DefSetUnion),
	% (get_entry_input_vars(Name, _In, Out)->   %Temporary fix won't work for multiple entry points.
	%  filter_args_aux2(DefSetUnion, Out, DefSetUnion2),!
	% ;
	%     DefSetUnion2=[],!
	% ),
	asserta_fact(block_def(Name,DefSetUnion)),
	block_def_set(Blocks).


filter_args(0, Name, Args, ArgsNew2):-    %filter output args
	entry_assert(Name,_),
	filter_args_aux(Args,ArgsNew),   %remove stack variables
	(get_entry_input_vars(Name, _In, Out)->
	 filter_args_aux2(ArgsNew, Out, ArgsNew2)
	 ;
	    ArgsNew2=ArgsNew
	).
% remove only stack arguments from function call input arguments.
filter_args(2, Name, Args, ArgsNew):-    %filter input args
	entry_assert(Name,_),
	filter_args_aux(Args,ArgsNew).   %remove stack variables


filter_args(1, Name, Args, ArgsNew2):-    %filter input args
	entry_assert(Name,_),
	filter_args_aux(Args,ArgsNew),   %remove stack variables
	(get_entry_input_vars(Name, In, _Out)->
	 filter_args_aux2(ArgsNew, In, ArgsNew2)
	 ;
	    ArgsNew2=ArgsNew
	).
filter_args(_, Name, Args, Args):-
	\+ entry_assert(Name,_).


filter_args_aux2([], _Out, []).

filter_args_aux2([A|Args], Out, Rest):-
	\+member(A,Out),
	filter_args_aux2(Args,Out,Rest).

filter_args_aux2([A|Args], Out, [A|Rest]):-
	member(A,Out),
	filter_args_aux2(Args,Out,Rest).



filter_args_aux([],[]).
filter_args_aux([A|Args],ArgsNew):-
	atom_concat('sp',_A,A),
	filter_args_aux(Args,ArgsNew).

filter_args_aux([A|Args],[A|ArgsNew]):-
	\+atom_concat('sp',_A,A),
	filter_args_aux(Args,ArgsNew).


getUnionOfDefs_From_Block(_Name, [],C,C).
getUnionOfDefs_From_Block(Name, [branch(_X,_Y)|Ns],C,D):-
	getUnionOfDefs_From_Block(Name, Ns,C,D).
% getUnionOfDefs_From_Block(Name, [calll(_X, _CallCounter)|Ns],C,D):-
% 	getUnionOfDefs_From_Block(Name, Ns,C,D).
getUnionOfDefs_From_Block(Name, [jmp(_X)|Ns],C,D):-
	getUnionOfDefs_From_Block(Name, Ns,C,D).
getUnionOfDefs_From_Block(Name, [Node|Ns], CurrSet, DefSetUnion):-
	def(% Name-
	   Node,DefSet1),
	sort(DefSet1, DefSet),
	ord_union(DefSet,CurrSet,CurrSet1),
	getUnionOfDefs_From_Block(Name, Ns,CurrSet1,DefSetUnion).
	

%%%%%%%%%%%%%%%%  Block-CFG  %%%%%%%%%%%%%%%%%%%

:- export(block_cfg/1).
block_cfg(Bcfg):-
 	getAllBlocksWithNames(Blocks),
	getVerticesEdges(Blocks,Vertices,Edges),
	vertices_edges_to_ugraph(Vertices,Edges,Bcfg).

%:- export(getVerticesEdges/3).
getVerticesEdges([],[],[]).
getVerticesEdges([(Name,Block)|Blocks],Vertices,Edges):-
	Vertices=[Name|Vertices1],
	getEdges(Name, Block, Edges1),
	getVerticesEdges(Blocks,Vertices1,Edges2),
	append(Edges1,Edges2,Edges).
	
getEdges(_Name,[],[]).
getEdges(Name,[branch(NameB1,NameB2)|Bs],Edges):-
	Edges=[Name-NameB1,Name-NameB2|Edges1],!,
	getEdges(Name,Bs, Edges1).



% %call instruction
% getEdges(Name,[B|Bs],Edges):-
% 	getInstrFromProgram(B,Instr),
% 	inst_interpret(Instr,B,[FunctionName|_Operands], _DefSet, _RefSet, bl,_Type),
% 	Edges=[Name-FunctionName|Edges1],!,
% 	getEdges(Name,Bs, Edges1).

%call instruction
% getEdges(Name,[calll(NameB, _CallCounter)|Bs],Edges):-
% 	Edges=[Name-NameB|Edges1],!,
% 	getEdges(Name,Bs, Edges1).



getEdges(Name,[jmp(NameB)|Bs],Edges):-
	Edges=[Name-NameB|Edges1],!,
	getEdges(Name,Bs, Edges1).

getEdges(Name,[_|Bs],Edges):-
	getEdges(Name,Bs, Edges).

% block_cfg_aux([(BlockName,Block)|Bs],Bcfg):-
% 	neighborBlocks(Block,Neighbors),
	

	

%%%%%%%%%%%%%%% connect blocks %%%%%%%%%%%%%%%%%%


%  [
% (b11,['07:','10:','11:','12:',call(b1111)]),
% (b1111,['01:','02:','03:','04:','05:','06:',call(b11)]),
% (b11,['08:','09:']),
% (b1,['13:','14:','15:'])
% ] ? 
% Add call nodes for all call sites in all blocks in database
% addCallNode:-
% 	findall(X,callEdge2(X),Xs),
% 	addCallNode_aux(Xs).

% addCallNode_aux([]).
% addCallNode_aux([N-Called-_Return|Rest]):-
% 	retrieve_block_containing_node1(Called, CalledBlock),
% 	retrieve_block_containing_node1(N, CalleeBlock),
% 	block(CalledBlockName, CalledBlock),
% 	insertCallNode(CalleeBlock, N, CalledBlockName, NewCalleeBlock),
% 	data_facts:retract_fact(block(Name1,CalleeBlock)),
% 	data_facts:asserta_fact(block(Name1,NewCalleeBlock)),
% 	addCallNode_aux(Rest).

% insertCallNode([], _N, _CalledBlockName, []).
% insertCallNode([N|Ns], N, CalledBlockName, [N, calll(CalledBlockName, CallCounter)|Ns]):-
% 	get_call_counter(CallCounter).

% insertCallNode([N1|Ns], N, CalledBlockName, [N1|Ns1]):-
% 	N1\=N,
% 	insertCallNode(Ns, N, CalledBlockName, Ns1).
	
	
%connectBlocks([B|Rest], Out):-
%:- export(retrieve_block_containing_node/2).	
retrieve_block_containing_node1(N,B):-
	block(_Name,B),
	member(N,B),!.

callEdge2(X):-
	callEdge(N,Called,Return),
	X=N-Called-Return.

% merge the return block with the call block. It happens when the branch is bl
%:- export(mergeBlocks/1).
% mergeBlocks(G):-
% 	findall(X,callEdge2(X),Xs),
% 	mergeBlocks_aux(Xs,G).
% %TOBE
% mergeBlocks_aux([],_G).
% mergeBlocks_aux([N-_Called-Return|Rest],G):-
% 	retrieve_block_containing_node1(N,CalleeBlock),
% 	retrieve_block_containing_node1(Return,ReturnedToBlock),
% 	mergeBlocks_(CalleeBlock,ReturnedToBlock,_B3,G),
% 	mergeBlocks_aux(Rest,G).
	
% %TOBE
% %:-export(mergeBlocks_/4).	
% mergeBlocks_(B1,B2,B3,Graph):-
% 	%check if last instr of B1 is a proc-call and returns to first instr of B2
% 	append(_X,[N|[]],B1),
% 	callEdge(N,Called,Return),
% 	Rest1=B2,
% 	Rest1=[Return|_RestX],
% 	%check if first instr of B2 has >1 inEdges
% 	inEdges(Return, Graph,  InEdgeSources),
% 	(InEdgeSources=[_X,_Y|_Rest2]->
% 	 %more than 1 inedges case
% 	 true;                             % Do nothing
% 	    %less than 2 inedges case
% 	    retrieve_block_containing_node1(Called,CalledBlock),
% 	    block(Name,CalledBlock),
% 	    get_call_counter(CallCounter),
% 	    CallEntry=..[calll|[Name, CallCounter]],
% 	    append(B1,[CallEntry|[]],B1_),
% 	    append(B1_,Rest1,B3),
% 	    data_facts:retract_fact(block(Name1,B1)),
% 	    data_facts:retract_fact(block(_,B2)),
% 	    data_facts:asserta_fact(block(Name1,B3))
% 	).


% add jmp/branch instructions at the end of each block.
%:- export(addJmpBranchNodes/1).
addJmpBranchNodes(G):-
	%findall(Block,block(_Name,Block),Blocks),
	getAllBlocksWithNames(Blocks),
	addJmpBranchNodes_aux(Blocks, G).

%asserta_fact(branch_info_table(BlockName, V1, Branch, Arg, A, V2)). % A=If Jump succesfull, V2= If jump is not successfull
addJmpBranchNodes_aux([],_G).
%TOBE
addJmpBranchNodes_aux([(Name,B)|Blocks],G):-
	append(_X,[N|[]],B), %get the last node
	outEdges(N,G,Targets),
	(Targets=[X,Y|[]]->
	 retrieve_block_containing_node1(X, B1),
	 retrieve_block_containing_node1(Y, B2),
	 addBranchNode((Name,B),B1,B2)
	;
	    Targets=[X|[]]->
	    retrieve_block_containing_node1(X, B1),
	    addJmpNode(B,B1)
	    ;
	     true
	),
	addJmpBranchNodes_aux(Blocks,G).
% branch_info(X,B1,Y,B2):-
% 	branch_info_table(_,bf,Arg,X,Y).
% branch_info(X,B1,Y,B2):-
% 	branch_info_table(_,bf,Arg,Y,X).
% branch_info(X,B1,Y,B2):-
% 	branch_info_table(_,bt,Arg,X,Y).
% branch_info(X,B1,Y,B2):-
% 	branch_info_table(_,bt,Arg,Y,X).
addJmpNode(B, B1):-
	block(NameB1, B1),	
	block(NameB, B),
	append(B, [jmp(NameB1)], B1_New),
	data_facts:retract_fact(block(NameB, B)),
	data_facts:asserta_fact(block(NameB,B1_New)).
%add a branch node from the end of B1 to B2.
:- data branch_info/2.
addBranchNode((Name, B), B1,B2):-
	block(NameB1, B1),	
	block(NameB2, B2),
	append(B, [branch(NameB1,NameB2)], B1_New),
	retractall_fact(branch_info(Name, _)),
	asserta_fact(branch_info(Name, branch(NameB1, NameB2))),
	retract_fact(block(NameB, B)),
	asserta_fact(block(NameB,B1_New)).



%%%%%%%%%%%%%%% block construction %%%%%%%%%%%%%%

%:- use_module(library(dynamic)).
%:- dynamic processed_node/1.
%processed_node(default).

:- data blockName/1.
blockName(b).
% generates a new block name with every new call by concatenating '1' with 'b'. Naive approach



getNewBlockName(Label, [A|_Block]):-
	get_label(A, Label), !.
%	entry_assert(Label,_), !.
getNewBlockName(B, _):-
	data_facts:retract_fact(blockName(Counter)),
	C is Counter + 1,
	atom_number(Cn, C),
	atom_concat(b, Cn, B),
	data_facts:asserta_fact(blockName(C)).

% getNewBlockName(B, _):-
% 	data_facts:retract_fact(blockName(Name)),
% 	atom_concat(Name,'1',B),
% 	data_facts:asserta_fact(blockName(B)).


get_label(A,Label):-
	data_facts:current_fact(program(Program)),
	reverse(Program,Program1),
	get_label_aux(Program1, A, Label).

get_label_aux([instr(Label,A,_I,_Format,_Operands)|_P], A, Label):-
	entry_assert(Label,_), !.

get_label_aux([instr(_Label,A,_I,_Format,_Operands)|P], A, L):-
	get_proc_label(P,L).

get_label_aux([instr(_Label,A1,_I,_Format,_Operands)|P], A, L):-
	\+A1==A,
	get_label_aux(P, A, L).

get_label_aux([proc_label(_Label)|P], A, L):-
	get_label_aux(P, A, L).

get_proc_label([proc_label(Label)|_P], Label):-
	entry_assert(Label,_), !.
get_proc_label([proc_label(_Label)|P], Label):-
	get_proc_label(P, Label).
	
	

% creates database entries into block(Name, Block).  Where 'Name' is
% an automatically generated name of each block and 'Block' is a list
% of instruction identifiers from database Program(X).
blockGraph(Graph, Blocks):-
	findall(Label, entry_assert(Label, _), Labels),
	l2a_map(L2A),
	blockGraph_aux(Labels, L2A, Graph, Blocks).

blockGraph_aux([],_L2A, _Graph, []).
blockGraph_aux([L|Ls], L2A, Graph, Blocks):-
	member(l2a(L, First), L2A),
	construct_blockGraph(First, Graph, B1),
	blockGraph_aux(Ls, L2A, Graph, B2),
	append(B1, B2, Blocks).
	

construct_blockGraph(First,Graph,Blocks):-
	push(First),                        %First= starting node of the graph
 	blockGraph_(Graph, Blocks,[]).

blockGraph_(_Graph,[],_Visited):-
	is_empty.
blockGraph_(Graph,[Block|Blocks],Visited):-
	\+is_empty,
	pop(A),
	constructBlockInit(A,Graph,Block),
	append(_X,[LastInstr|[]],Block),
	neighbors(LastInstr, Graph, Neighbors),
	push_list(Neighbors,[A|Visited]),  % Push Neighbors to stack if not already visited.
	blockGraph_(Graph,Blocks,[A|Visited]),!, % To avoid infinit recursion
	getNewBlockName(Name, Block),
	data_facts:asserta_fact(block(Name,Block)).

push_list([A|Rest],Visited):-
	member(A,Visited),
	push_list(Rest,Visited).
push_list([A|Rest],Visited):-
	\+member(A,Visited),
	(exist_in_stack(A)->
	 true
	;
	    push(A)
	),!,
	push_list(Rest,Visited).
push_list([],_V).

% push_new_block_starters(Block,Graph):-
% 	append(_X,[LastInstr|[]],Block),
% 	neighbors(LastInstr, Graph, Neighbors),
		
% First node
constructBlockInit(A,Graph,[A]):-  
	outEdges(A , Graph,  OutEdgeSources),
	OutEdgeSources=[_X,_Y|_Xs], !.             %if first node is a branch.

constructBlockInit(A,Graph,[A|Block]):-
	inEdges(A, Graph, InEdges),
	\+InEdges=[_X],
	\+InEdges=[],                           %if first node has more than 1 inedge.
	outEdges(A , Graph,  OutEdgeSources),
	OutEdgeSources=[B],
	constructBlock(B,Graph,Block ), !.

constructBlockInit(A,Graph,[A|Block]):- 
	outEdges(A , Graph,  OutEdgeSources),
	OutEdgeSources=[B],
	constructBlock(B,Graph,Block ), !.
constructBlockInit(A,_Graph,[A]).

% 2nd and onwards nodes.
constructBlock(A,Graph,[A|Block] ):-
	inEdges(A, Graph, InEdges),
	InEdges=[_X],
	outEdges(A, Graph, Neighbors),
	( Neighbors=[Y]-> %If not a branch
	   constructBlock(Y,Graph,Block);
	    %If a brach
	   Block=[] %include the last instruction = branching
	),!.
constructBlock(A,Graph,[] ):-
	inEdges(A, Graph, InEdges),
	\+InEdges=[_X].

constructBlock(_,_,[]).
%%%%%%%%%%%%%%%%%%%%% All node reachability %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%:- export(graphh/1).
% graphh(G):-
% 	vertices_edges_to_ugraph([k,f,g,h,l],[f-g,f-l,g-h,h-k,k-g],G).
:- export(reachable/2).
%:- export(reachableF/2).
:- export(block_output_vars/2).
%:- export(block_def/2).
%:- export(block_ref/2).
:- data block_output_vars/2.
:- data reachable/2.
:- data reachableF/2. 
%:-export(fcleanup/0).
fcleanup:-
	retractall_fact(reachable(_,_)),
	retractall_fact(reachableF(_,_)),
	retractall_fact(block_output_vars(_,_)).

%%%%%%%%%%%%%%% output variables

blockOutputVars1([]).
blockOutputVars1([(Name,_B)|Bs]):-
	get_entry_input_vars(Name, _In, Out),!,
	asserta_fact(block_output_vars(Name, Out)),!,
	blockOutputVars1(Bs).
blockOutputVars1([(Name,_B)|Bs]):-
	asserta_fact(block_output_vars(Name, [])),!,
	blockOutputVars1(Bs).


% outputVar(Block)= [ {i=1 to n}Union block_def(block_i) where block_i=reachable(block) ] Union block_def(block)
%:- export(blockOutputVars/0).
blockOutputVars(B):-
	retractall_fact(block_output_vars(_,_)),
	% getAllBlocksWithNames(B),
	blockOutputVars_aux(B).

blockOutputVars_aux([]).
blockOutputVars_aux([(Name,_B)|Bs]):-
	block_def(Name,DefSet),                            %Vars def/modify in block
	reachable(Name,ReachableBlocks),                   %reachable blocks from current block 'Name'
	getDefSet(ReachableBlocks,DefSet1),                %Get the def/modify set of the reachable blocks
	ord_union(DefSet,DefSet1,BlockOutput),             %Union all 
	filter_args(0, Name, BlockOutput, BlockOutput1),
	asserta_fact(block_output_vars(Name, BlockOutput1)),
	blockOutputVars_aux(Bs).
getDefSet([],[]).
getDefSet([N|Ns], Set):-
	block_def(N,DefSet),
	getDefSet(Ns, Set1),
	ord_union(DefSet, Set1, Set).
%%%%%%%%%%%%%%% update output variables
:-export(updateBlockOutputVars/1).
updateBlockOutputVars(Blocks):-
	block_cfg(G),
	transpose(G, TG),
	updateBlockOutputVars_aux(Blocks, TG).

updateBlockOutputVars_aux(Blocks, Block_cfg_transpose):-
	updateBlockOutputVars_aux1(Blocks, Block_cfg_transpose, Updated),
	(Updated>0->
	 updateBlockOutputVars_aux(Blocks, Block_cfg_transpose)
	;
	    true
	).
updateBlockOutputVars_aux(_Blocks,_).

updateBlockOutputVars_aux1([], _TG, 0).
updateBlockOutputVars_aux1([(Name,_B)|Bs], TG, Updated):-
	entry_assert(Name,_),
	updateBlockOutputVars_aux1(Bs, TG, Updated).
	
updateBlockOutputVars_aux1([(Name,_B)|Bs], TG, Updated):-
	\+entry_assert(Name,_),
	neighbors(Name, TG, Predecesor_Blocks),
	output_vars_union(Predecesor_Blocks, ParentsOutVars),
	block_output_vars(Name, OutputVars),
	ord_union(OutputVars, ParentsOutVars, Union),
	(OutputVars==Union->
	 Updated_curr=0
	;
	    retractall_fact(block_output_vars(Name,_)),
	    asserta_fact(block_output_vars(Name, Union)),
	    Updated_curr=1
	),
	updateBlockOutputVars_aux1(Bs, TG, Updated_aux),
	Updated is Updated_curr+Updated_aux.
	    

output_vars_union([], []).
output_vars_union([B|Bs], OutVarsUnion):-
	block_output_vars(B, Out1),
	output_vars_union(Bs, OutVarsUnion1),
	ord_union(Out1, OutVarsUnion1, OutVarsUnion).
	
	

	
%%%%%%%%%%%%%%% input variables
%:- export(def_so_far/2).
:- data def_so_far/2.
:- export(block_input_vars/2).
:- data block_input_vars/2.

%getBlockForm('~/transform/fact_assembly.txt',Blocks), inst_def_ref, getAllBlocksWithNames(B),block_def_set(B),reachableVerticesBlock,blockOutputVars,computeDefUnion,block_ref_set(B).
%:- export(blockInputVars/0).
blockInputVars:-
	retractall_fact(block_input_vars(_,_)),
	getAllBlocksWithNames(B),
	blockInputVars_aux(B),
	blockInputVars_branch_arg(B).

%fixpoint
blockInputVars_branch_arg(B):-
	blockInputVars_branch_arg_aux(B,Flag),
	(Flag=1->
	 blockInputVars_branch_arg(B);
	    true
	).
	
blockInputVars_branch_arg_aux([],0).
blockInputVars_branch_arg_aux([(Name,_B)|Bs],Flag):-
	entry_assert(Name,_),
	blockInputVars_branch_arg_aux(Bs, Flag).
	
blockInputVars_branch_arg_aux([(Name,_B)|Bs],Flag):-
	\+entry_assert(Name,_),
	block_input_vars(Name, BlockInput),   %inferred for block 'Name'
	reachable(Name,ReachableBlocks),
	getInputSet(ReachableBlocks,InputSet),%inferred for the reachable blocks from block 'Name'
	ord_subtract(InputSet, BlockInput, Diff),
	(Diff=[]-> %if no difference 
	 blockInputVars_branch_arg_aux(Bs, Flag);
	    Flag=1,
	    ord_union(BlockInput, Diff, BlockInput_New),
	    retractall_fact(block_input_vars(Name, _)),
	    asserta_fact(block_input_vars(Name, BlockInput_New))
	    
	).
getInputSet([],[]). 
getInputSet([N|Ns], Set):-
	block_input_vars(N,RefSet),
	getInputSet(Ns, Set1),
	ord_union(RefSet, Set1, Set).

blockInputVars_aux([]).
blockInputVars_aux([(Name,B)|Bs]):-
	block_ref(Name,RefSet), 
	reachable(Name,ReachableBlocks),
	getRefSet(ReachableBlocks,RefSet1),
	ord_union(RefSet,RefSet1,BlockInput1),
	filter_args(1, Name, BlockInput1, BlockInput_),
	%filter_args(2, Name, BlockInput1, BlockInput_),
	get_branch_var(Name, B, BVar),
	ord_union(BlockInput_, BVar, BlockInput),
	asserta_fact(block_input_vars(Name, BlockInput)),
	blockInputVars_aux(Bs).
getRefSet([],[]). 
getRefSet([N|Ns], Set):-
	block_ref(N,RefSet),
	getRefSet(Ns, Set1),
	ord_union(RefSet, Set1, Set).


%getBlockForm('~/transform/fact_assembly.txt',Blocks), inst_def_ref, getAllBlocksWithNames(B),block_def_set(B),reachableVerticesBlock,blockOutputVars,computeDefUnion.
block_ref_set(B):-
	%getAllBlocksWithNames(B),
	block_ref_aux(B).

block_ref_aux([]).
block_ref_aux([(Name,[B|Bs])|Blocks]):-
	ref(% Name-
	   B,Ref1),
	sort(Ref1, Ref),
	block_ref_aux1(Name, [B|Bs],Ref,RefSetUnion1),
	filter_args(1, Name, RefSetUnion1, RefSetUnion),
	asserta_fact(block_ref(Name,RefSetUnion)),
	block_ref_aux(Blocks).

block_ref_aux1(_Name, [_N|[]],C,C).

% block_ref_aux1(Name, [_N,calll(X, _CallCounter)|Ns],CurrSet,RefUnion):-
% 	block_ref_aux1(Name, [calll(X, _CallCounter)|Ns],CurrSet,RefUnion).
block_ref_aux1(Name, [_N,jmp(X)|Ns],CurrSet,RefUnion):-
	block_ref_aux1(Name, [jmp(X)|Ns],CurrSet,RefUnion).

block_ref_aux1(Name, [_N,branch(X,Y)|Ns],CurrSet,RefUnion):-
	block_ref_aux1(Name, [branch(X,Y)|Ns],CurrSet,RefUnion).

% block_ref_aux1(Name, [calll(X, _CallCounter),N2|Ns],CurrSet,RefUnion):-
% 	ref(% Name-
% 	   N2,RefSet1),
% 	sort(RefSet1, RefSet),
% 	block_output_vars(X,OutputVars),
% 	ord_subtract(RefSet,OutputVars,Ref),
% 	ord_union(Ref,CurrSet,CurrSet1),
% 	block_ref_aux1(Name, [N2|Ns],CurrSet1,RefUnion).
	
block_ref_aux1(Name, [jmp(X),N2|Ns],CurrSet,RefUnion):-
	ref(% Name-
	       N2,RefSet1),
	       write('not possible'),nl,
	sort(RefSet1, RefSet),
	block_output_vars(X,OutputVars),
	ord_subtract(RefSet,OutputVars,Ref),
	ord_union(Ref,CurrSet,CurrSet1),
	block_ref_aux1(Name, [N2|Ns],CurrSet1,RefUnion).

block_ref_aux1(Name, [branch(X,Y),N2|Ns],CurrSet,RefUnion):-
	ref(% Name-
	   N2,RefSet1),
	sort(RefSet1, RefSet),
	block_output_vars(X,OutputVarsX),
	block_output_vars(Y,OutputVarsY),
	ord_union(OutputVarsX,OutputVarsY,OutputVars),
	ord_subtract(RefSet,OutputVars,Ref),
	ord_union(Ref,CurrSet,CurrSet1),
	block_ref_aux1(Name, [N2|Ns],CurrSet1,RefUnion).

block_ref_aux1(Name, [N1,N2|Ns],CurrSet,RefUnion):-
	ref(% Name-
	   N2,RefSet1),
	sort(RefSet1, RefSet),
	def_so_far(Name-N1,DefUnion),
	ord_subtract(RefSet,DefUnion,Ref),
	ord_union(Ref,CurrSet,CurrSet1),
	block_ref_aux1(Name, [N2|Ns],CurrSet1,RefUnion).


%:- export(computeDefUnion/0).
%getBlockForm('~/transform/fact_assembly.txt',Blocks), inst_def_ref, getAllBlocksWithNames(B),block_def_set(B),reachableVerticesBlock,blockOutputVars.
computeDefUnion(B):-
	retractall_fact(def_so_far(_,_)),
	% getAllBlocksWithNames(B),		
	computeDefUnion_aux(B).

computeDefUnion_aux([]).
%[(Name,_B)|Bs]
computeDefUnion_aux([(Name,B)|Bs]):-
	def_till_node(Name, B,[]), % the empty list will be set of input variables in case we know in advance
	computeDefUnion_aux(Bs).
% Def set at node n is the union of all the def sets of previous nodes.
def_till_node(_Name, [],_OldDefSet).

% def_till_node(Name, [calll(X, _CallCounter)|Ns],OldDefSet):-
% 	block_output_vars(X,OutputVar),
% 	ord_union(OutputVar,OldDefSet,NewDefSet),
% 	asserta_fact(def_so_far(Name-calll(X, _CallCounter),NewDefSet)),
% 	def_till_node(Name, Ns,NewDefSet).

def_till_node(Name, [jmp(X)|Ns],OldDefSet):-
	block_output_vars(X,OutputVar),
	ord_union(OutputVar,OldDefSet,NewDefSet),
	asserta_fact(def_so_far(Name-jmp(X),NewDefSet)),
	def_till_node(Name, Ns,NewDefSet).

def_till_node(Name, [branch(X,Y)|Ns],OldDefSet):-
	block_output_vars(X,OutputVarX),
	block_output_vars(Y,OutputVarY),
	ord_union(OutputVarX,OutputVarY,NewDefSetXY),
	ord_union(NewDefSetXY,OldDefSet,NewDefSet),
	asserta_fact(def_so_far(Name-branch(X,Y),NewDefSet)),
	def_till_node(Name, Ns,NewDefSet).

def_till_node(Name, [N|Ns],OldDefSet):-
	def(% Name-
	   N,DefSet1),
	sort(DefSet1, DefSet),
	ord_union(DefSet,OldDefSet,NewDefSet),
	asserta_fact(def_so_far(Name-N,NewDefSet)),
	def_till_node(Name, Ns, NewDefSet).
	


%%%%%%%%%%%%%%%
%Each node is attached with a set, which is intially all the neighbours of the node
% reachableFixpoint(Node) and reachable(Node) both are equal to Neighbors(Node)
initialize_fixpoint(_Graph,[]).
initialize_fixpoint(Graph,[V|Vs]):-
	neighbors(V, Graph, Ns),
	sort(Ns,Ns1),
	asserta_fact(reachable(V,Ns1)),
	asserta_fact(reachableF(V,Ns1)),
	initialize_fixpoint(Graph, Vs).

% true if reachableFixpoint=reachable...  if no difference between the previous and current state then return true
isStateSame([]).
isStateSame([V|Vs]):-
	reachable(V, Set2),
	reachableF(V, Set1),
	ord_union_diff(Set1, Set2, _Union, Diff),
	(Diff=[]-> %if 
	 isStateSame(Vs);
	    false
	).


% getBlockForm('~/transform/fact_assembly.txt',Blocks), inst_def_ref, getAllBlocksWithNames(B),block_def_set(B),reachableVerticesBlock.

%One iteration, reachable/2 is updated	
computeNewState(_G,[]).
computeNewState(G, [V|Vs]):-
	reachable(V,Set),
	computeNewState_aux(Set, G, NewSet1), %Take each neighbor k of V and union its neighbours to current neighbor set of V.
	ord_union(Set,NewSet1,NewSet),
	data_facts:retractall_fact(reachable(V,_)),
	data_facts:asserta_fact(reachable(V,NewSet)),
	computeNewState(G,Vs).

computeNewState_aux([],_G,[]).
computeNewState_aux([K|Ks], G, NewSet):-
	neighbors(K,G,K_Ns),
	computeNewState_aux(Ks,G,NewSet1),
	ord_union(K_Ns,NewSet1,NewSet).
	

% Compute set of reachable vertices from each vertex from V in graph G.
%:- export(fixpoint_start/2).
fixpoint_start(G,V):-
	fcleanup,
	initialize_fixpoint(G,V), %reachableFixpoint(Node) and reachable(Node) both are equal to Neighbors(Node)
	fixpoint_aux(G,V).

fixpoint_aux(G,V):-
	computeNewState(G,V),     %One iteration, reachable/2 is updated	
	(isStateSame(V)->
	 true;
	    recordFixpointState(V), %reachableFixpoint = reachable	    
	    fixpoint_aux(G,V)
	).
%reachableFixpoint = reachable	    

recordFixpointState([]).
recordFixpointState([V|Vs]):-
	reachable(V,Set),
	retractall_fact(reachableF(V,_)),
	asserta_fact(reachableF(V,Set)),
	recordFixpointState(Vs).


reachableVerticesBlock:-
	block_cfg(Bcfg),
	vertices(Bcfg,V),
	fixpoint_start(Bcfg,V).

%%%%%%%%%%%%%%% Graph construction %%%%%%%%%%%%%%%%%
%:- export(generate_graph/3).

generate_graph(G,First):-
	compute_vertices(V),
	append([First|[]],_Rest,V),
	compute_edges(V,Edges),
	vertices_edges_to_ugraph(V,Edges,G),
	data_facts:asserta_fact(cfg(G)).
	

%p:-
%	assert(f(s)),parse:varr1("d","d",_R).


% control_graph:-
% 	 file_read('~/transform/fact_assembly.txt',Program),
% 	 generate_graph(Program,Graph).

% generate_graph(Program,Graph):-
% 	compute_vertices(Program,Vertices,[],L2A,A2L).
%:- export(compute_edges/2).
compute_edges(V,Edges):-
	data_facts:current_fact(program(Program)),
	data_facts:current_fact(l2a_map(L2A)),
	compute_edges_aux(Program,V,Edges,L2A).

compute_edges_aux(_Program, [],[],_L2A).
compute_edges_aux(_Program, [_V1|[]],[],_L2A).
compute_edges_aux(Program, [V1,V2|Vertices],Edges,L2A):-
	member(instr(_Label,V1,Instr,Format,Operands),Program),
	next_instr(V1,Instr,Format,Operands,V2,L2A,NextInstrs),
	buildEdges(V1,NextInstrs,Edges1),
	compute_edges_aux(Program, [V2|Vertices], Edges2,L2A),
	append(Edges1,Edges2,Edges).

buildEdges(_V1,[],[]).

buildEdges(V1,[N|NextInstrs],[E|Edges]):-
	E=V1-N,
	buildEdges(V1,NextInstrs,Edges).%,
	%identifyReturnEdge(V1,N).

% identifyReturnEdge(V1,N):-
% 	data_facts:current_fact(program(Program)),
% 	member(instr(_Label,V1,Instr,Format,Operands),Program)

:- data branch_info_table/6.
:- export(branch_info_table/6).

% Following needs to be in interpreter.
% branch_info_table(V1, Branch, Arg, A, V2)
next_instr(V1,Branch, ru6, operands(Arg, L), V2, L2A, [A,V2]):- 
	member(Branch,[bf,bt]),
	atom_codes(L,LL),
	extract_label(L1,LL,_R),
	member(l2a(L1,A),L2A),!,
	asserta_fact(branch_info_table(_BlockName, V1, Branch, Arg, A, V2)). % A=If Jump succesfull, V2= If jump is not successfull, blockName filled later

next_instr(_V1,bu, u6, operands(L), _V2, L2A, [A]):-
	atom_codes(L,LL),
	extract_label(L1,LL,_R),
	member(l2a(L1,A),L2A),!.

next_instr(V1,bl, u10, operands(L), V2, L2A, [V2]):-
	atom_codes(L,LL),
	extract_label(L1,LL,_R),
	member(l2a(L1,A),L2A),!,
	data_facts:asserta_fact(call_edge(V1-A)),
	data_facts:asserta_fact(return_edge(V1-V2)).
next_instr(_V1,retsp, _, _, _V2, _L2A, []):-!.
next_instr(_V1,_,_,_,V2,_,[V2]).
	
	



%:- export(compute_vertices/1).
compute_vertices(V):-
	data_facts:current_fact(program(Program)),
	compute_vertices_aux(Program,V,[],L2A,_A2L),
	data_facts:asserta_fact(l2a_map(L2A)).
	

compute_vertices_aux([],[],_L,[],[]).

compute_vertices_aux([proc_label(L)|Rest],Vertices,Labels,L2A,_A2L):-
	compute_vertices_aux(Rest,Vertices,[L|Labels],L2A,_A2L).

compute_vertices_aux([instr(L,A,_I,_F,_Os)|Rest],[A|Vertices],Labels,L2A,_A2L):-
	(L=''->
	 l2a(A,Labels,L2A_);                    %compute label to address mappings
	 l2a(A,[L|Labels],L2A_)                     %compute label to address mappings
	),                   
	append(L2A_,L2A_2,L2A),
	compute_vertices_aux(Rest,Vertices,[],L2A_2,_A2L_2).

	
% a2l(_A,[],[]).
% %l2a(A,['',R],L2A):-l2a(A,R,L2A).
% a2l(A,[L|Rest],A2L):-
% 	A2L=[a2l(A,L)|A2L_],
% 	l2a(A,Rest,A2L_).

l2a(_A,[],[]).
%l2a(A,['',R],L2A):-l2a(A,R,L2A).
l2a(A,[L|Rest],L2A):-
	L2A=[l2a(L,A)|L2A_],
	l2a(A,Rest,L2A_).
	

:- doc(preds2unfold/2, "Takes as input @var{ModuleName} the name of
the current Module and returns @var{PredsWithArity} the blocks to
preserve duing unfolding in a particular format: 'ModuleName:BlockName'/arity").
%preds2unfold(_, ['findMin:b3'/5]):-!.
preds2unfold(ModuleName, PredsWithArity):-
	block_cfg(Bcfg),
	preds2unfold_aux(Bcfg, Preds),
	preds_arity(Preds,ModuleName, PredsWithArity).

preds2unfold_aux([],[]).
preds2unfold_aux([B-[_B1,_B2]|Xs],[B|Preds]):-!,
	preds2unfold_aux(Xs,Preds).
preds2unfold_aux([_X|Xs],Preds):-
	preds2unfold_aux(Xs,Preds).

preds_arity([],_, []).

preds_arity([P|Preds],ModuleName, [P1/A|PredsWithArity]):-
	atom_concat(ModuleName,':', P_),
	atom_concat(P_,P, P1),
	block_input_vars(P,InputVars),
	block_output_vars(P,OutputVars),
	length(InputVars, L1),
	length(OutputVars, L2),
	A is L1+L2,
	preds_arity(Preds, ModuleName, PredsWithArity).